from itertools import zip_longest
from pathlib import Path
from typing import List

import typer
from pynmrstar import Entry, Loop, Saveframe
from strenum import StrEnum

from nef_pipelines.lib.nef_lib import (
    NEF_PIPELINES_PREFIX,
    NEF_RELAXATION_VERSION,
    UNUSED,
    SelectionType,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames,
    select_frames_by_name,
)
from nef_pipelines.lib.peak_lib import frame_to_peaks
from nef_pipelines.lib.shift_lib import IntensityMeasurementType
from nef_pipelines.lib.util import exit_error, parse_comma_separated_options
from nef_pipelines.tools.fit.fit_lib import (
    _exit_if_spectra_are_missing,
    _get_atoms_and_values,
    _get_spectra_by_series_variable,
)
from nef_pipelines.tools.series import series_app

NAMESPACE = NEF_PIPELINES_PREFIX


class SeriesSelectionMethod(StrEnum):
    """Selection types for selecting frames"""

    ASSIGNMENT = "name"
    POSITION = "position"
    TAG = "tag"


@series_app.command()
def table(
    input: Path = typer.Option(
        None,
        "-i",
        "--in",
        metavar="NEF-FILE",
        help="read NEF data from a file instead of stdin",
    ),
    exact: bool = typer.Option(
        False,
        "--exact",
        help="use exact matching of frame names",
    ),
    # series_type: SeriesSelectionMethod = typer.Argument(
    #     SelectionType.ASSIGNMENT,
    #     help='how to select groups of peaks in the series frames',
    # ),
    frames_selectors: List[str] = typer.Argument(
        None,
        help="series frames to build data tables for",
    ),
    intensity_measurement_type: IntensityMeasurementType = typer.Option(
        IntensityMeasurementType.HEIGHT,
        "--measurement-type",
        help="measurement type to use for intensity values",
    ),
    force: bool = typer.Option(
        False,
        "--force",
        help="force overwriting of the table if one alread exists",
    ),
    add_atom_names: bool = typer.Option(
        False,
        "--add-atom-names",
        help="decorate the table with the names of the atoms in the peaks",
    ),
):
    """- build a NEF data series from a set of spectra [alpha]"""

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    frame_selectors = parse_comma_separated_options(frames_selectors)

    _exit_if_no_frame_selectors(frame_selectors)

    frames = _select_series_frames(entry, frames_selectors, exact=exact)

    entry = pipe(
        entry,
        frames,
        intensity_measurement_type,
        SeriesSelectionMethod.ASSIGNMENT,
        force,
        add_atom_names,
    )

    print(entry)


# TODO: add checks if there are peaks with duplicate atom names and exit as error
def pipe(
    entry: Entry,
    series_frames: List[str],
    intensity_measurement_type: str,
    series_selection_method: SeriesSelectionMethod,
    force: bool,
    add_atom_names: bool,
) -> Entry:

    for series_frame in series_frames:
        relaxation_loop_name = series_frame.name
        series_data_loop = _build_series_data_loop(
            series_frame,
            intensity_measurement_type,
            entry,
            relaxation_loop_name,
            add_atom_names,
        )

        _exit_if_frame_has_loop_unless_force(
            series_frame, series_data_loop, entry, force
        )

        series_frame.add_loop(series_data_loop)
    return entry


def _exit_if_frame_has_loop_unless_force(series_frame, series_data_loop, entry, force):
    if series_data_loop.category in series_frame and not force:
        msg = f"""
                in the entry {entry.id}
                the frame {series_frame.name} already has a loop with the category {series_data_loop.category}

                to force addition of the loop use the --force option
            """
        exit_error(msg)


def _build_series_data_loop(
    series_frame,
    intensity_measurement_type,
    entry,
    relaxation_loop_name,
    add_atom_names,
):

    series_experiment_loop = (
        series_frame[f"_{NAMESPACE}_series_experiment"]
        if f"_{NAMESPACE}_series_experiment" in series_frame
        else None
    )

    spectra_by_times_and_indices = _get_spectra_by_series_variable(
        entry, series_experiment_loop
    )

    _exit_if_spectra_are_missing(spectra_by_times_and_indices, series_frame.name)

    peaks_by_times_and_indices = {
        key: frame_to_peaks(spectrum_frame)
        for key, spectrum_frame in spectra_by_times_and_indices.items()
    }

    atoms_and_values = _get_atoms_and_values(
        peaks_by_times_and_indices, intensity_measurement_type
    )

    series_data_loop = Loop.from_scratch(f"{NAMESPACE}_series_data")

    tags = [
        "version",
        "nmr_spectrum_id",
        "peak_id",
        "variable_value",
        "variable_error",
        "value",
        "value_error",
        "relaxation_list_id",
        "data_id",
    ]

    additional_tags = []
    if add_atom_names:
        max_num_atoms = max([len(atoms) for atoms in atoms_and_values.keys()])
        additional_tags = [
            f"nefpls_chain_code_{i} nefpls_sequence_code_{i} nefpls_residue_name_{i} nefpls_atom_name_{i}".split()
            for i in range(1, max_num_atoms + 1)
        ]
    tags.extend(additional_tags)

    series_data_loop.add_tag(tags)

    for group_index, (atoms, values) in enumerate(atoms_and_values.items(), start=1):
        zipper = zip_longest(
            values.spectra,
            values.peak_ids,
            values.variable_values,
            values.values,
            fillvalue=UNUSED,
        )

        for spectrum, peak_id, series_value, value in zipper:
            data = {
                "version": NEF_RELAXATION_VERSION,
                "nmr_spectrum_id": spectrum,
                "peak_id": peak_id,
                "variable_value": series_value,
                "variable_error": UNUSED,
                "value": value,
                "value_error": UNUSED,
                "relaxation_list_id": f"{relaxation_loop_name}_fitted",
                "data_id": group_index,
            }

            if add_atom_names:
                for i, atom in enumerate(atoms, start=1):
                    data.update(
                        {
                            f"nefpls_chain_code_{i}": atom.residue.chain_code,
                            f"nefpls_sequence_code_{i}": atom.residue.sequence_code,
                            f"nefpls_residue_name_{i}": atom.residue.residue_name,
                            f"nefpls_atom_name_{i}": atom.atom_name,
                        }
                    )

            series_data_loop.add_data(
                [
                    data,
                ]
            )

    return series_data_loop


def _select_series_frames(
    entry: Entry, frame_selectors: List[str], exact: bool
) -> List[Saveframe]:
    raw_series_frames = select_frames(
        entry, f"{NAMESPACE}_series_list", SelectionType.CATEGORY
    )

    return select_frames_by_name(raw_series_frames, frame_selectors, exact=exact)


def _exit_if_no_frame_selectors(frame_selectors):
    if len(frame_selectors) == 0:
        msg = "you must select some frames!"
        exit_error(msg)


# def _guess_experiment_type(name):
#     upper_name = name.upper()
#     for experiment_name in NAME_TO_EXPERIMENT_TYPE:
#         if experiment_name in upper_name:
#             result = NAME_TO_EXPERIMENT_TYPE[upper_name]
#             break
#         else:
#             result = "unknown"
#     return result
#
#
# def _guess_series_name(entry, frames_and_timings):
#     result = None
#     for relaxation_type in NAME_TO_EXPERIMENT_TYPE:
#         do_break = False
#         for name, _ in frames_and_timings.keys():
#             name = name.upper()
#             if relaxation_type in name.upper():
#                 start_index = name.index(relaxation_type)
#                 end_index = start_index + len(relaxation_type)
#                 found_relaxation_type = name[start_index:end_index]
#
#                 result = found_relaxation_type
#                 do_break = True
#                 break
#         if do_break:
#             break
#
#     if not result:
#         result = "unknown_type"
#
#     bad_frame_name = False
#     for frame in entry:
#         if frame.name == result:
#             bad_frame_name = True
#             break
#
#     if bad_frame_name:
#         base_result = result
#         for i in range(1, 100):
#             result = f"series`{i}`"
#             bad_frame_name = False
#             for frame in entry:
#                 if frame.name == result:
#                     bad_frame_name = True
#                     break
#
#     if bad_frame_name:
#         _exit_if_too_many_clashing_frame_names(bad_frame_name, base_result)
#
#     return result
#
#
# def _exit_if_too_many_clashing_frame_names(bad_frame_name, base_result):
#     if bad_frame_name:
#         msg = f"""
#             after searching the series of names {base_result}`1` to {base_result}`100` i could not find a name
#             that was not already in the entry, choose a name on the command line, you have way too much data!
#             more serious report this as a bug...
#         """
#         exit_error(msg)
#
#
# def _get_timings_and_units_for_frames(
#     frames, frame_selectors, timings, input_unit, display_parsed_values
# ):
#
#     if timings:
#         timings_and_units = _parse_timings(timings)
#         units = [timing_and_unit[1] for timing_and_unit in timings_and_units]
#
#         _exit_if_input_unit_defined_and_any_unit_parsed(
#             input_unit, units, timings_and_units
#         )
#
#         if input_unit:
#             timings_and_units = [
#                 (timing, input_unit) for timing, _ in timings_and_units
#             ]
#
#         _exit_error_it_number_of_timings_doesnt_match_number_of_frames(
#             frames, timings_and_units
#         )
#
#         frame_timings_and_units = {
#             (frame.name, id(frame)): timing_and_unit
#             for frame, timing_and_unit in zip(frames, timings_and_units)
#         }
#     else:
#         _exit_if_frame_selector_has_no_timing_selector(frame_selectors)
#         frame_timings_and_units = _parse_timings_from_frames_or_exit(
#             frames, frame_selectors[0], display_parsed_values
#         )
#
#     return frame_timings_and_units
#
#
# def _exit_if_input_unit_defined_and_any_unit_parsed(
#     input_unit, units, timings_and_units
# ):
#     if input_unit and any(unit != "" for unit in units):
#         units_and_timings_strings = [
#             f"{unit}{timing}" for unit, timing in timings_and_units if unit
#         ]
#         msg = f"""
#                 you cannot provide a unit on the command line [--unit] and units in the timings [--timings]
#                 at the same time command line unit was {input_unit} and the units and timings were
#                 {strings_to_tabulated_terminal_sensitive(units_and_timings_strings)}
#
#                 """
#         exit_error(msg)
#
#
# def _exit_error_it_number_of_timings_doesnt_match_number_of_frames(
#     frames, timings_and_units
# ):
#     if len(timings_and_units) != len(frames):
#         msg = f"the number of timings {len(timings_and_units)} does not match the number of frames {len(frames)}"
#         exit_error(msg)
#
#
# def _exit_if_frame_selector_has_no_timing_selector(frame_selectors):
#     if len(frame_selectors) != 1 or "{var}" not in frame_selectors[0]:
#         NEWLINE = "\n"
#         msg = f"""
#                     no timings provided to select timings you must have a single frame selector with
#                     a {{var}} placeholder i got {len(frame_selectors)} frame selectors and they were:
#                     {NEWLINE.join(frame_selectors)}
#                 """
#         exit_error(msg)
#
# def _exit_if_some_frames_are_not_spectra(non_spectrum_frame_names):
#     msg = f"""
#             some of the selected frames are not spectrum frames:
#             {non_spectrum_frame_names}
#             """
#     exit_error(msg)
#
# def _parse_timings_from_frames_or_exit(
#     save_frames: List[Saveframe], frame_selector: str, display_parsed_values: bool
# ) -> Dict[Tuple[str, int], List[Tuple[Union[int, float, bool], str]]]:
#     if "{var}" not in frame_selector:
#         msg = f"no timing selector placeholder {{var}} provided in frame selector {frame_selector}"
#         exit_error(msg)
#
#     frame_selector = frame_selector.replace("{var}", "{var}")
#     frame_selector = frame_selector.replace("*", "{:opt}")
#
#     parser = parse_compile(frame_selector)
#
#     parsed_values = []
#
#     frame_value_and_units = {}
#     for frame in save_frames:
#         frame_key = frame.name, id(frame)
#         result = parser.search(frame.name)
#
#         value, unit = None, None
#         if result and "var" in result:
#             raw_value = result["var"]
#             raw_value = raw_value.replace("_", ".")
#             raw_value = raw_value.replace("-", ".")
#
#             value, unit = _parse_value_and_unit(raw_value)
#
#         if display_parsed_values:
#             parsed_values.append([frame.name, raw_value, value, unit])
#
#         if value is None:
#             _exit_if_no_variable_found(frame, frame_selector)
#
#         if not (is_float(value) or value.lower() in "on off true false yes no"):
#             msg = f"the variable {value} in frame {frame.name} is not a number or a boolean"
#             exit_error(msg)
#
#         frame_value_and_units[frame_key] = value, unit
#
#     if display_parsed_values:
#         print(
#             tabulate(parsed_values, headers=["spectrum", "raw value", "value", "unit"]),
#             file=sys.stderr,
#         )
#         sys.exit(0)
#
#     return frame_value_and_units
#
#
# def _exit_if_no_variable_found(frame, frame_selector):
#     msg = f"no variable found in frame {frame.name} with the selector {frame_selector}"
#     exit_error(msg)
#
#
# def _parse_timings(timings: List[str]) -> List[Tuple[Union[int, float, bool], str]]:
#     timings = parse_comma_separated_options(timings)
#     parsed_timings = []
#     for timing in timings:
#         value, unit = _parse_value_and_unit(timing)
#         parsed_timings.append((value, unit))
#     return parsed_timings
#
#
# def _parse_value_and_unit(value: str) -> Tuple[Union[bool, int, float, None]]:
#
#     unit = value.lstrip(string.digits + ".eE")
#     len_digits = len(value) - len(unit)
#     value = value[:len_digits]
#
#
#     lower_unit = unit.lower()
#     if value and unit:
#         pass
#     elif not value and lower_unit in STR_TO_SATURATION_STATE:
#         unit = "bool"
#         value = STR_TO_SATURATION_STATE[lower_unit]
#     elif not unit:
#         if is_int(value):
#             value = int(value)
#         elif is_float(value):
#             value = float(value)
#     else:
#         unit =  None
#         value =  None
#
#     return value, unit
#

#
# def _ensure_units_consistent_and_get_unit_and_timings_or_exit(frames_timings_and_units):
#
#     _exit_units_not_consistent(frames_timings_and_units)
#
#     _exit_if_numbers_and_bools_are_mixed(frames_timings_and_units)
#
#     float_counter = Counter()
#     int_counter = Counter()
#     for frame, (value, unit) in frames_timings_and_units.items():
#         if isinstance(value, int):
#             int_counter[value] += 1
#         elif isinstance(value, float):
#             float_counter[value] += 1
#
#     if int_counter and float_counter:
#         for frame_key, (value, unit) in frames_timings_and_units.items():
#             value = float(value)
#
#             frames_timings_and_units[frame_key] = value, unit
#
#     unit = list(frames_timings_and_units.values())[0][1]
#     frames_and_timings = {
#         frame_key: value for frame_key, (value, _) in frames_timings_and_units.items()
#     }
#
#     return frames_and_timings, unit
#
#
# def _exit_if_numbers_and_bools_are_mixed(frame_timings_and_units):
#
#     counter_numbers = Counter()
#     couter_booleans = Counter()
#
#     for frame, (value, unit) in frame_timings_and_units.items():
#         if isinstance(value, bool):
#             couter_booleans[value] += 1
#         elif is_float(value):
#             counter_numbers[value] += 1
#
#     if counter_numbers and couter_booleans:
#         numbers_strings = ", ".join(counter_numbers.keys())
#         booleans_strings = ", ".join(couter_booleans.keys())
#         msg = f"""
#                 you cannot have both numbers and booleans in the values
#                 the booleans were
#                 {booleans_strings}
#                 the numbers were
#                 {numbers_strings}
#             """
#         exit_error(msg)
#
#
# def _exit_units_not_consistent(frame_timings_and_units):
#
#     unit_counter = Counter()
#     for frame, (value, unit) in frame_timings_and_units.items():
#         unit_counter[unit] += 1
#
#     if len(unit_counter) > 1:
#         unit_strings = ", ".join(unit_counter.keys())
#         msg = f"""
#             the units of the timings are not consistent, the units found were
#             {unit_strings}
#             """
#         exit_error(msg)
#
#
# def _exit_if_units_are_not_compatible(frame_timings_and_units):
#     units = {unit for timing, unit in frame_timings_and_units.values()}
#     if len(units) > 1:
#         units_and_timings_strings = [
#             f"{unit}{timing}" for timing, unit in frame_timings_and_units.values()
#         ]
#         msg = f"""
#                 the units of the timings are not compatible they are {units}
#                 the timings and units were
#                 {strings_to_tabulated_terminal_sensitive(units_and_timings_strings)}
#                 """
#         exit_error(msg)



from nef_pipelines.lib.test_lib import assert_lines_match
from nef_pipelines.transcoders.nmrpipe.nmrpipe_lib import format_pipe_sequence


def test_format_pipe_sequence():
    TEST_DATA = ["A"] * 110

    BLOCK = "AAAAAAAAAA"
    EXPECTED = f"""\
        DATA SEQUENCE {BLOCK} {BLOCK} {BLOCK} {BLOCK} {BLOCK} {BLOCK} {BLOCK} {BLOCK} {BLOCK} {BLOCK}
        DATA SEQUENCE {BLOCK}
    """

    lines = format_pipe_sequence(TEST_DATA)
    result = "\n".join(lines)

    assert_lines_match(result, EXPECTED)



from textwrap import dedent

import typer

from nef_pipelines.lib.structures import (
    AtomLabel,
    SequenceResidue,
    ShiftData,
    ShiftList,
)
from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    read_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.nmrview.importers.shifts import shifts
from nef_pipelines.transcoders.nmrview.nmrview_lib import parse_shifts

SHIFTS_NMRVIEW = "nef_chemical_shift_list_nmrview"


app = typer.Typer()
app.command()(shifts)


def test_lib_parse_shifts():

    EXPECTED = ShiftList(
        shifts=[
            ShiftData(
                atom=AtomLabel(
                    SequenceResidue(
                        chain_code="A", sequence_code=1, residue_name="ASP"
                    ),
                    atom_name="CA",
                ),
                value=52.00,
                value_uncertainty=None,
            ),
            ShiftData(
                atom=AtomLabel(
                    SequenceResidue(
                        chain_code="A", sequence_code=1, residue_name="ASP"
                    ),
                    atom_name="HA",
                ),
                value=4.220,
                value_uncertainty=None,
            ),
            ShiftData(
                atom=AtomLabel(
                    SequenceResidue(
                        chain_code="A", sequence_code=2, residue_name="VAL"
                    ),
                    atom_name="CG2",
                ),
                value=19.300,
                value_uncertainty=None,
            ),
            ShiftData(
                atom=AtomLabel(
                    SequenceResidue(
                        chain_code="A", sequence_code=2, residue_name="VAL"
                    ),
                    atom_name="HG21",
                ),
                value=0.814,
                value_uncertainty=None,
            ),
            ShiftData(
                atom=AtomLabel(
                    SequenceResidue(
                        chain_code="A", sequence_code=2, residue_name="VAL"
                    ),
                    atom_name="HG22",
                ),
                value=0.814,
                value_uncertainty=None,
            ),
            ShiftData(
                atom=AtomLabel(
                    SequenceResidue(
                        chain_code="A", sequence_code=3, residue_name="GLN"
                    ),
                    atom_name="N",
                ),
                value=125.058,
                value_uncertainty=None,
            ),
        ]
    )

    test_data = """\
        1.CA      52.000 1
        1.HA       4.220 1
        2.CG2     19.300 1
        2.HG21     0.814 1
        2.HG22     0.814 1
        3.N      125.058 1
    """

    chain_seqid_to_type = {("A", 1): "ASP", ("A", 2): "VAL", ("A", 3): "GLN"}

    test_data = dedent(test_data)

    shifts = parse_shifts(test_data.split("\n"), chain_seqid_to_type)

    assert shifts == EXPECTED


EXPECTED_PPM_OUT_SHORT = """\
    save_nef_chemical_shift_list_nmrview
        _nef_chemical_shift_list.sf_category   nef_chemical_shift_list
        _nef_chemical_shift_list.sf_framecode  nef_chemical_shift_list_nmrview

        loop_
            _nef_chemical_shift.chain_code
            _nef_chemical_shift.sequence_code
            _nef_chemical_shift.residue_name
            _nef_chemical_shift.atom_name
            _nef_chemical_shift.value
            _nef_chemical_shift.value_uncertainty
            _nef_chemical_shift.element
            _nef_chemical_shift.isotope_number

            A   1   ASP   CA     52.0      .   .   .
            A   1   ASP   HA     4.22      .   .   .
            A   2   VAL   CG2    19.3      .   .   .
            A   2   VAL   HG21   0.814     .   .   .
            A   2   VAL   HG22   0.814     .   .   .
            A   3   GLN   N      125.058   .   .   .

        stop_

    save_
"""


# # noinspection PyUnusedLocal
def test_ppm_out_short_no_sequence():

    STREAM = read_test_data("header.nef", __file__)

    path = path_in_test_data(__file__, "ppm_short.out")
    result = run_and_report(app, [path], expected_exit_code=1, input=STREAM)

    assert "ERROR" in result.stdout
    assert "did you read a sequence?" in result.stdout


# noinspection PyUnusedLocal
def test_ppm_out_short():

    STREAM = read_test_data("ppm_short_seq.nef", __file__)

    path = path_in_test_data(__file__, "ppm_short.out")

    result = run_and_report(app, [path], input=STREAM)

    mol_sys_result = isolate_frame(result.stdout, SHIFTS_NMRVIEW)

    assert_lines_match(EXPECTED_PPM_OUT_SHORT, mol_sys_result)



import sys
from fnmatch import fnmatch
from typing import List

import typer
from ordered_set import OrderedSet
from treelib import Tree
from typer import Context

from nef_pipelines.lib.util import parse_comma_separated_options
from nef_pipelines.tools.help import help_app

VERBOSE_HELP = """\
    display more verbose information about NEF-Pipelines and its environment,
    [ignored with the --version option]"
"""


# noinspection PyUnusedLocal
@help_app.command()
def commands(
    context: Context,
    matchers: List[str] = typer.Argument(
        None,
        help="select commands to display multiple items, wildcards and comma separated lists are allowed",
    ),
    # tree: bool = typer.Option(False, "--table", help="display the tree in a table format"),
    # python: bool = typer.Option(False, "-v", "--python", help='show commands with python interfaces'),
):
    "- display and filter the help for the NEF-Pipelines commands [alpha]"

    matchers = parse_comma_separated_options(matchers)
    if not matchers:
        matchers = [
            "*",
        ]

    tree = cmd(context, matchers)

    if len(tree) != 0:
        print(tree.show(stdout=False), file=sys.stderr)
        print("key: [X] has a python function [P]ipe / [C]md", file=sys.stderr)
        print("     [α] alpha feature", file=sys.stderr)


def _get_function_help(data):
    return data.callback.__doc__ if data.callback and data.callback.__doc__ else ""


def cmd(context: Context, matchers: List[str]) -> Tree:
    app = context.find_root()
    root_command = app.command

    tree_elems = {path: object for path, object in _walk_tree(root_command, "nef")}

    raw_filtered_tree = []
    for path in tree_elems:
        matches = [fnmatch(".".join(path), f"*{matcher}*") for matcher in matchers]
        if all(matches):
            raw_filtered_tree.append(path)

    filtered_tree = OrderedSet()
    for path in raw_filtered_tree:
        for elem in _path_components(path):
            if elem not in filtered_tree:
                filtered_tree.add(elem)

    tree = Tree()
    for path in filtered_tree:
        node_name = path[-1]
        node_path = ".".join(path)
        parent_path = ".".join(path[:-1])
        parent_path = None if not parent_path else parent_path
        data = tree_elems[path]

        programming_commands = _get_python_commands(data)
        programming_decorators = [
            f"[{command.__name__[0].upper()}]" for command in programming_commands
        ]
        help = _get_function_help(data)
        if "[alpha]" in help:
            programming_decorators.append("[α]")

        programming_decorators = " ".join(programming_decorators)

        node_title = f"{node_name} {programming_decorators}"

        tree.create_node(node_title, node_path, parent=parent_path, data=data)

    return tree


def _get_python_commands(data):
    commands = []
    if data.callback:
        module_name = data.callback.__module__
        module = sys.modules[module_name]
        for name in "pipe", "cmd":
            if hasattr(module, name):
                commands.append(getattr(module, name))
    return commands


def _path_components(path):
    result = []
    for elem in path:
        result.append(elem)
        yield tuple(result)


def _walk_tree(root, root_name):

    def objwalk(obj, path=()):
        path = [*path]

        yield path, obj
        if hasattr(obj, "commands"):
            for child in obj.commands.values():
                for child_path, child in objwalk(child, [*path, child.name]):
                    yield child_path, child

    for path, elem in objwalk(root, [root_name]):
        yield tuple(path), elem



import itertools
import re
import string
from collections import Counter
from dataclasses import replace
from pathlib import Path
from typing import List

import typer
from fyeah import f
from pynmrstar import Entry

from nef_pipelines.lib.isotope_lib import CODE_TO_ISOTOPE, GAMMA_RATIOS
from nef_pipelines.lib.nef_lib import (
    add_frames_to_entry,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.util import (
    NEWLINE,
    STDIN,
    exit_error,
    is_float,
    parse_comma_separated_options,
)
from nef_pipelines.transcoders.deep import import_app
from nef_pipelines.transcoders.nmrpipe.nmrpipe_lib import (
    get_gdb_columns,
    read_db_file_records,
    read_peak_file,
    select_data_records,
)
from nef_pipelines.transcoders.nmrview.importers.peaks import _create_spectrum_frame

DEEP_PEAK_EXPECTED_FIELDS = (
    "INDEX X_AXIS X_PPM  XW X1 X3  HEIGHT ASS CONFIDENCE POINTER".split()
)

app = typer.Typer()

# TODO should check peaks against sequence


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def peaks(
    chain_codes: str = typer.Option(
        "A", "--chain", help="chain code", metavar="<chain-code>"
    ),
    # filter_noise: bool = typer.Option(True, help="remove peaks labelled as noise"),
    in_file: Path = typer.Option(
        STDIN, "-i", "--in", help="file to read nef data from", metavar="<NEF-FILE>"
    ),
    spectrometer_frequencies: List[float] = typer.Option(
        ..., "--spectrometer-frequency", help="the 1H frequencies of the spectrometers"
    ),
    entry_name: str = typer.Option(
        "nmrpipe", "-e", "--entry", help="entry name", metavar="<entry-name>"
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="input peak files", metavar="<peak-file.xpk>"
    ),
):
    """convert deep peak file <DEEP>.tab files to NEF [alpha]"""

    number_files = len(file_names)
    if not chain_codes:
        chain_codes = ["A"] * number_files

    if len(chain_codes) == 1:
        chain_codes = chain_codes * number_files

    spectrometer_frequencies = parse_comma_separated_options(spectrometer_frequencies)

    number_spectrometer_frequencies = len(spectrometer_frequencies)
    if (
        number_spectrometer_frequencies > 1
        and number_files != number_spectrometer_frequencies
    ):
        _exit_lenghts_spectrometer_frequencies_and_files_dont_match(
            file_names, spectrometer_frequencies
        )
    elif number_spectrometer_frequencies == 1 and number_files > 1:
        spectrometer_frequencies = spectrometer_frequencies * number_files

    chain_codes = parse_comma_separated_options(chain_codes)

    _exit_if_num_chains_and_files_dont_match(chain_codes, file_names)

    entry = read_or_create_entry_exit_error_on_bad_file(in_file, entry_name=entry_name)

    entry = pipe(entry, file_names, chain_codes, spectrometer_frequencies)

    print(entry)


def pipe(
    entry: Entry,
    file_names: List[Path],
    chain_codes: List[str],
    spectrometer_frequencies: List[float],
) -> Entry:

    gdb_files = _read_gdb_files(file_names)
    peak_list_map = _read_deep_peaks(gdb_files, chain_codes)

    deep_sweep_widths_map, deep_isotopes_map = _read_deep_sweep_widths_and_isotopes(
        gdb_files
    )

    frame_name_template = "{file_name}"

    for i, (path, peak_list) in enumerate(peak_list_map.items()):

        if peak_list.peak_list_data.sweep_widths is None:
            new_peak_list_data = replace(peak_list.peak_list_data, sweep_widths=[])
            peak_list = replace(peak_list, peak_list_data=new_peak_list_data)
            peak_list_map[path] = peak_list

        deep_sweep_widths = deep_sweep_widths_map[path]
        for axis_label in peak_list.peak_list_data.axis_labels:
            peak_list.peak_list_data.sweep_widths.append(deep_sweep_widths[axis_label])

        deep_isotopes = deep_isotopes_map[path]
        for axis_label in peak_list.peak_list_data.axis_labels:
            axis_isotope = deep_isotopes[axis_label]
            isotope = CODE_TO_ISOTOPE[axis_isotope]
            gamma_ratio = GAMMA_RATIOS[isotope]
            peak_list_map[path].peak_list_data.spectrometer_frequencies.append(
                spectrometer_frequencies[i] * gamma_ratio
            )

    frame_names = _create_entry_names_from_template_if_required(
        frame_name_template, [], file_names
    )

    frames = []
    for i, (peak_list, frame_name, chain_code) in enumerate(
        zip(peak_list_map.values(), frame_names, chain_codes), start=1
    ):
        frames.append(_create_spectrum_frame(frame_name, peak_list, chain_code))

    # TODO when adding frames they need to be disambiguated if there are name clashes
    return add_frames_to_entry(entry, frames)


# TODO move much of this to utilities
def _exit_if_num_chains_and_files_dont_match(chain_codes, file_names):
    if len(chain_codes) != len(file_names):
        msg = f"""
            the number of chain codes {len(chain_codes)} does not match number of files {len(file_names)}
            the chain codes are {', '.join(chain_codes)}
            the files are
            {NEWLINE.join(file_names)}
        """
        exit_error(msg)


def _disambiguate_names(names):
    name_count = Counter()
    result = []
    for name in names:
        if name in name_count:
            result.append(f"{name}_{name_count[name]}")
        else:
            result.append(name)
        name_count[name] += 1
    return result


def _remove_repeated_underscores(s):
    return re.sub("_+", "_", s)


def _create_entry_names_from_template_if_required(
    entry_name_template, entry_names, file_names
):
    new_entry_names = []
    for file_name, entry_name in itertools.zip_longest(file_names, entry_names):
        if entry_name:
            new_entry_names.append(entry_name)
        else:
            file_name = str(Path(file_name).stem)
            numbers_and_letters = string.ascii_lowercase + string.digits
            file_name = "".join(  # noqa: F841
                [
                    letter if letter in numbers_and_letters else "_"
                    for letter in file_name
                ]
            )

            new_name = f(entry_name_template)

            new_name = _remove_repeated_underscores(new_name)

            new_entry_names.append(new_name)

    return _disambiguate_names(new_entry_names)


def _read_gdb_files(file_names):
    results = []
    for file_name in file_names:
        with open(file_name) as file_h:
            gdb_file = read_db_file_records(file_h, file_name=file_name)

        _check_is_peak_file_or_exit(gdb_file)

        results.append(gdb_file)

    return results


def _read_deep_peaks(gdb_files, chain_codes):
    results = {}
    for gdb_file, chain_code in zip(gdb_files, chain_codes):

        results[gdb_file.name] = read_peak_file(gdb_file, chain_code)

    return results


def _check_is_peak_file_or_exit(gdb_file):
    columns = set(get_gdb_columns(gdb_file))
    data_records = select_data_records(gdb_file, "*_AXIS")
    data_record_names = [data_record.values[0] for data_record in data_records]

    if not _check_is_deep_file(gdb_file):
        if set(DEEP_PEAK_EXPECTED_FIELDS) != set(columns):
            msg = f"""\
                    this gdb file doesn't appear to contain all the columns expected for an deep peak file
                    expected: {','.join(DEEP_PEAK_EXPECTED_FIELDS)}
                    got {','.join(columns & set(DEEP_PEAK_EXPECTED_FIELDS))}
                    file name: {gdb_file.name}
                    """
            exit_error(msg)
        else:
            msg = f"""\
                    this gdb file doesn't appear to contain all the columns expected for an deep peak file
                    I expected DATA records begining X_AXIS, Y_AXIS etc
                    got {', '.join(data_record_names)}
                    file_name {gdb_file.name}
                   """
            exit_error(msg)


def _check_is_deep_file(gdb_file):

    columns = set(get_gdb_columns(gdb_file))
    expected_fields = set(DEEP_PEAK_EXPECTED_FIELDS)

    data_records = select_data_records(gdb_file, "*_AXIS")
    return expected_fields.issubset(columns) and data_records


def _report_bad_sweep_width_line(data_record, message):
    line_info = data_record.line_info
    msg = f"""
                        Couldn't read axis DATA records of the form
                        DATA *_AXIS  <isotope> <left-point> <right-point> <left-ppm> <right-ppm> such as

                        DATA  X_AXIS 1H           1   779   11.303ppm    5.149ppm

                        because {message}
                        at line {line_info.line_no} in {line_info.file_name}

                        the line was

                        {line_info.line}
                    """
    exit_error(msg)


def _position_to_ppm(position):
    result = None
    if position.endswith("ppm"):
        position = position[: -len("ppm")]
        if is_float(position):
            result = float(position)

    return result


def _parse_axis(axis):

    result = None

    axis = axis.split("_")
    if len(axis) == 2:
        result = axis[0]

    return result


def _read_deep_sweep_widths_and_isotopes(gdb_files):
    sweep_widths_result = {}
    isotopes_result = {}

    for gdb_file in gdb_files:
        data_records = select_data_records(gdb_file, "*_AXIS")
        for data_record in data_records:
            try:
                values = data_record.values
                axis, isotope, _, _, raw_right_ppm, raw_left_ppm = values
            except Exception:
                msg = f"wrong number of records (got {len(values)} expected 6)"
                _report_bad_sweep_width_line(data_record, msg)

            axis = _parse_axis(axis)

            if not axis:
                msg = f"can't parse axis field ({axis}) should be of the form *_AXIS where * is one of X,Y,Z,A"
                _report_bad_sweep_width_line(data_record, msg)

            left_ppm = _position_to_ppm(raw_left_ppm)
            right_ppm = _position_to_ppm(raw_right_ppm)

            if not left_ppm:
                _report_bad_sweep_width_line(
                    data_record, f" left ppm position ({raw_left_ppm}) is bad"
                )
            if not right_ppm:
                _report_bad_sweep_width_line(
                    data_record, f" right ppm position ({raw_right_ppm}) is bad"
                )

            sweep_width = right_ppm - left_ppm
            sweep_widths_result.setdefault(gdb_file.name, {})[axis] = sweep_width

            isotopes_result.setdefault(gdb_file.name, {})[axis] = isotope

    return sweep_widths_result, isotopes_result


def _exit_lenghts_spectrometer_frequencies_and_files_dont_match(
    file_names, spectrometer_frequencies
):
    pass



import csv
import re
import string
from enum import auto
from io import TextIOWrapper
from pathlib import Path
from textwrap import dedent
from typing import Dict, List

import typer
from ordered_set import OrderedSet
from pynmrstar import Entry
from strenum import StrEnum
from tabulate import tabulate

from nef_pipelines.lib.isotope_lib import ATOM_TO_ISOTOPE
from nef_pipelines.lib.nef_lib import (
    UNUSED,
    add_frames_to_entry,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.peak_lib import peaks_to_frame
from nef_pipelines.lib.sequence_lib import (
    get_residue_name_from_lookup,
    sequence_from_entry,
)
from nef_pipelines.lib.structures import (
    AtomLabel,
    LineInfo,
    NewPeak,
    SequenceResidue,
    ShiftData,
)
from nef_pipelines.lib.util import (
    STDIN,
    chunks,
    exit_error,
    flatten,
    get_display_file_name,
    is_float,
    is_int,
    parse_comma_separated_options,
)
from nef_pipelines.transcoders.csv import import_app

app = typer.Typer()

HELP_FOR_FRAME_NAME = "name for the frame, note white spaces will be replace by _"
HELP_FOR_FORMATS = """\
    CSV like formats that can be read [CSV: comma separated variables, TSV tab separated variables,
    SSV space separated variables, AUTO read file and guess format from contents]
"""


class CsvLikeFormats(StrEnum):
    CSV = auto()
    TSV = auto()
    SSV = auto()
    AUTO = auto()
    csv = auto()
    tsv = auto()
    ssv = auto()
    auto = auto()


INDEX = "index"
PEAK_ID = "peak_id"

CHAIN_CODE_STUB = "chain_code_"
SEQUENCE_CODE_STUB_ = "sequence_code_"
RESIDUE_NAME_STUB = "residue_name_"
ATOM_NAME_STUB = "atom_name_"
ATOM_CODES = flatten(
    [
        f"{CHAIN_CODE_STUB}{i} {SEQUENCE_CODE_STUB_}{i} {RESIDUE_NAME_STUB}{i} {ATOM_NAME_STUB}{i}".split()
        for i in range(1, 16)
    ]
)
POSITION_STUB = "position_"
POSITION_UNCERTAINTY_STUB = "position_uncertainty_"
POSITION_CODES = [f"{POSITION_STUB}{i}".split() for i in range(1, 16)]
POSITION_UNCERTAINTY_CODES = [
    f"{POSITION_UNCERTAINTY_STUB}{i}".split() for i in range(1, 16)
]
VOLUME = "volume"
VOLUME_UNCERTAINTY = "volume_uncertainty"
HEIGHT = "height"
HEIGHT_UNCERTAINTY = "height_uncertainty"

BASIC_HEADERS = flatten([PEAK_ID, *ATOM_CODES, *POSITION_CODES])
OPTIONAL_HEADERS = [
    VOLUME,
    VOLUME_UNCERTAINTY,
    HEIGHT,
    HEIGHT_UNCERTAINTY,
    *POSITION_UNCERTAINTY_CODES,
]
REQUIRED_HEADERS = [PEAK_ID, *ATOM_CODES[:4], *POSITION_CODES[0]]

COLUMN_SEPARATORS_MAY_HAVE_CHANGED = (
    "note: column separators shown here may different to those in the original file..."
)
DEFAULT_ATOMS_HELP = (
    "the atoms to use if atoms names aren't defined in the file "
    "This option can be used multiple twice to define the atoms or you can use a "
    "comma separated list"
)

DEFAULT_NUCLEI_HELP = (
    "nuclei to use for each dimension if not defined they are guessed from the assigments"
    "or an error is reported"
)


class CSVPeakListException(Exception):
    pass


class BadSequenceException(CSVPeakListException):
    pass


class BadDimensionException(CSVPeakListException):
    pass


class IncompatibleDimensionTypesException(CSVPeakListException):
    pass


@import_app.command(no_args_is_help=True)
def peaks(
    entry_input: Path = typer.Option(
        STDIN,
        metavar="<INPUT>",
        help="file to read NEF data from [stdin = -]",
    ),
    csv_format: CsvLikeFormats = typer.Option(
        CsvLikeFormats.AUTO, "-f", "--format", help=HELP_FOR_FORMATS
    ),
    csv_file_encoding: str = typer.Option(
        "utf-8-sig", "-e", "--encoding", help="encoding for the csv file"
    ),
    default_atoms: List[str] = typer.Option(
        [], "-a", "--atoms", help=DEFAULT_ATOMS_HELP
    ),
    dimension_nuclei: List[str] = typer.Option([], help=DEFAULT_NUCLEI_HELP),
    default_chain_code: str = typer.Option(
        "A",
        "-c",
        "--chain-code",
        help="default chain code to use if none is provided in the file",
    ),
    spectrometer_frequency: float = typer.Option(
        600.123456789, help="spectrometer frequency in MHz"
    ),
    csv_file: Path = typer.Argument(None, metavar="<CSV-FILE>"),
):
    default_atoms = parse_comma_separated_options(default_atoms)

    csv_format = csv_format.upper()

    entry = read_or_create_entry_exit_error_on_bad_file(entry_input)

    try:
        pipe(
            entry,
            default_chain_code,
            default_atoms,
            csv_file,
            csv_file_encoding,
            csv_format,
            dimension_nuclei,
            spectrometer_frequency,
        )
    except BadSequenceException as e:

        sequence_table = _tabulate_sequence(entry)
        _exit_bad_sequence(e, entry_input, sequence_table)

    except BadDimensionException as e:
        msg = f"peak list dimensions were not usable because {e}"
        exit_error(msg)

    except CSVPeakListException as e:
        exit_error(str(e))

    print(entry)


def _exit_bad_sequence(e, entry_input, sequence_table=None):
    msg = (
        f"{dedent(str(e))}\n"
        f"the source of the residue sequence was: {get_display_file_name(entry_input)}\n\n"
    )
    if not sequence_table:
        msg += "note: if you would like to see a complete input sequence use the verbose option"
    else:
        msg += "sequence was:\n\n"
        msg += sequence_table

    exit_error(msg)


def _guess_dimensions_if_not_defined_or_throw(
    peaks: List[NewPeak], input_dimensions: List[str]
) -> List[str]:
    results_by_dimension = {
        i: dimension for i, dimension in enumerate(input_dimensions)
    }

    guessed_dimensions = {}

    for peak_number, peak in enumerate(peaks, start=1):

        for dimension_index, shift in enumerate(peak.shifts):
            atom_name = shift.atom.atom_name

            if len(atom_name) == 0:
                atom_name = None

            atom_type = None
            if atom_name != UNUSED and atom_name is not None:
                atom_type = atom_name.strip()[0]

            if atom_type is not None and atom_type.lower() not in string.ascii_letters:
                atom_type = None

            guessed_dimension_isotopes = guessed_dimensions.setdefault(
                dimension_index, set()
            )
            if atom_type and atom_type in ATOM_TO_ISOTOPE:
                isotope_code = ATOM_TO_ISOTOPE[atom_type]
                guessed_dimension_isotopes.add(isotope_code)
            else:
                guessed_dimension_isotopes.add(UNUSED)

    for guessed_dimension_index in guessed_dimensions:
        if guessed_dimension_index not in results_by_dimension:

            guessed_dimension = guessed_dimensions[guessed_dimension_index]
            if len(guessed_dimension) > 1 and UNUSED in guessed_dimension:
                guessed_dimension.remove(UNUSED)

            if len(guessed_dimensions[guessed_dimension_index]) > 1:
                _raise_multiple_guessed_isotopes(
                    guessed_dimensions, guessed_dimension_index
                )

    for dimension_index in guessed_dimensions:
        if dimension_index not in results_by_dimension:
            results_by_dimension[dimension_index] = list(
                guessed_dimensions[dimension_index]
            )[0]

    max_dimension = max(results_by_dimension.keys())

    result = []
    for i in range(max_dimension + 1):
        result.append(results_by_dimension[i])

    return result


def _raise_multiple_guessed_isotopes(guessed_dimensions, guessed_dimension_index):
    dim_info = []
    for dimension_index in sorted(guessed_dimensions).keys():
        dim_data = f"{dimension_index}: {' '.join(guessed_dimensions.values())}"
        dim_info.append(dim_data)
    msg = f"""\
                    no isotope provided for dimension {guessed_dimension_index + 1} and multiple possibilities
                    found in atom names, please use explicit isotope codes using dimension-nuclei

                    deduced codes for dimensions:


                """
    msg = dedent(msg)
    msg += dim_info
    raise IncompatibleDimensionTypesException(msg)


def pipe(
    entry: Entry,
    chain_code: str,
    atoms: List[str],
    csv_file: Path,
    csv_file_encoding: str,
    csv_format: CsvLikeFormats,
    input_dimensions: List[str],
    spectrometer_frequency,
) -> Entry:
    encoding = {"encoding": csv_file_encoding}

    file_shifts = _parse_csv(
        csv_file, encoding, chain_code, atoms, csv_file, csv_format
    )

    dimensions = _guess_dimensions_if_not_defined_or_throw(
        file_shifts, input_dimensions
    )

    dimensions = [{"axis_code": dimension} for dimension in dimensions]

    frame = peaks_to_frame(file_shifts, dimensions, spectrometer_frequency)

    add_frames_to_entry(entry, [frame])

    return entry


def _parse_csv(
    csv_file: Path,
    encoding: Dict[str, str],
    default_chain_code: str,
    atoms: List[str],
    file_name: str,
    csv_format: str,
) -> List[NewPeak]:
    column_offsets = {}
    headers = [*BASIC_HEADERS, *OPTIONAL_HEADERS]

    peaks = []
    with open(csv_file, **encoding) as csv_fp:

        peak_reader = _get_csv_reader_for_format(csv_format, csv_fp, encoding)

        for i, row in enumerate(peak_reader):
            row = [elem.strip() for elem in row]

            line_info = LineInfo(file_name, i + 1, ", ".join(row))
            if i == 0:
                header_row = [elem.lower() for elem in row]

                _exit_if_headers_dont_look_right(row, i + 1, header_row, file_name)

                for header in headers:
                    if header in header_row:
                        column_offsets[header] = header_row.index(header)

            else:
                shifts = []
                for column_index in range(1, 16):
                    chain_code_header = f"chain_code_{column_index}"
                    if chain_code_header in column_offsets:
                        chain_code = row[column_offsets[chain_code_header]]
                    else:
                        chain_code = default_chain_code

                    sequence_code_header = f"sequence_code_{column_index}"
                    sequence_code = None
                    if sequence_code_header in column_offsets:
                        sequence_code = row[column_offsets[sequence_code_header]]

                    if not sequence_code:
                        continue

                    residue_name = extract_column_or_unused(
                        row, f"residue_name_{column_index}", column_offsets
                    )

                    atom_name = extract_column_or_unused(
                        row, f"atom_name_{column_index}", column_offsets
                    )

                    # TODO this is a required column note if its missing!
                    position = extract_column_or_unused(
                        row, f"position_{column_index}", column_offsets
                    )
                    exit_if_value_is_unused(
                        position, f"position_{column_index}", line_info
                    )
                    position = float(position)

                    position_uncertainty = extract_column_or_unused(
                        row, f"position_uncertainty_{column_index}", column_offsets
                    )
                    if position_uncertainty != UNUSED:
                        position_uncertainty = float(position_uncertainty)

                    sequence_residue = SequenceResidue(
                        chain_code, sequence_code, residue_name
                    )
                    atom_label = AtomLabel(sequence_residue, atom_name)
                    shift = ShiftData(atom_label, position, position_uncertainty)
                    shifts.append(shift)

                peak = NewPeak(shifts)
                peaks.append(peak)
    return peaks


def extract_column_or_unused(row, residue_name_header, column_offsets):
    if residue_name_header in column_offsets:
        residue_name = row[column_offsets[residue_name_header]]
    else:
        residue_name = UNUSED
    return residue_name


def exit_if_value_is_unused(value, header, line_info):
    name = header.rstrip(string.digits).rstrip("_")
    if value == UNUSED:
        msg = f"""\
            at line {line_info.line_no} in file {line_info.file_name} for column {header}
            all atom names in the peak must have a {name} [. is not permissible] i got {value}
            line was: {line_info.line}
            note: columns separators may not be the same as in your file
        """
        exit_error(msg)


def _exit_if_headers_dont_look_right(row: List[str], row_number, header_row, file_name):

    row_set = set(row)
    row_set -= {
        INDEX,
    }
    if not set(REQUIRED_HEADERS).issubset(set(row)):
        msg = f"""
            in the file {file_name} at the first row the headers for the file don't look right,
            as a minimum you should have

            {' '.join(REQUIRED_HEADERS)}

            the following were missing from the first row

            {' '.join(set(REQUIRED_HEADERS) - set(row))}

            which has the following headings

            {' '.join(row)}

        """
        exit_error(msg)

    remaining_headings = OrderedSet(row)
    remaining_headings -= {INDEX, PEAK_ID}

    for required_peak_info, required_shift_info in zip(
        chunks(ATOM_CODES, 4), POSITION_CODES
    ):

        if not set(required_peak_info).issubset(row_set) and not set(
            required_shift_info
        ).issubset(required_shift_info):
            remaining_headings -= set(required_peak_info)
            remaining_headings -= set(required_shift_info)
            msg = f"""
                in the file {file_name} at the first row not all of the headers that must be seen together are present
                it is expected that the following headers should be seen together

                {' '.join(required_peak_info)}

                and

                {' '.join(required_shift_info)}

                the remaining headings are



                {' '.join(remaining_headings)}

            """
            exit_error(msg)
        else:
            remaining_headings -= required_peak_info
            remaining_headings -= required_shift_info


def _exit_error_if_value_not_type(value, type, value_msg, row, row_number, file_name):
    TYPES = {int: (is_int, "integer"), float: (is_float, "float")}

    converter, type_name = TYPES[type]

    if not converter(value):
        msg = f"""\
            at line {row_number} in file {file_name} for the value {value}
            {value_msg}
            {', '.join(row)} [{COLUMN_SEPARATORS_MAY_HAVE_CHANGED}]
        """
        exit_error(msg)


def _get_csv_reader_for_format(csv_format, csv_fp, encoding):
    if csv_format == CsvLikeFormats.AUTO:
        dialect = csv.Sniffer().sniff(csv_fp.read(1024))
        csv_fp.seek(0)
        rdc_reader = csv.reader(csv_fp, dialect)
    elif csv_format == CsvLikeFormats.TSV:
        rdc_reader = csv.reader(csv_fp, delimiter="\t")
    elif csv_format == CsvLikeFormats.CSV:
        rdc_reader = csv.reader(csv_fp)
    else:
        lines = []
        for line in csv_fp:
            line = re.sub(r"\s+", "\t", line)
            lines.append(line.strip())
        lines = "\n".join(lines)
        csv_fp = TextIOWrapper(lines, encoding)
        rdc_reader = csv.reader(csv_fp, delimiter="\t")
    return rdc_reader


def _lookup_residue_name_or_exit(chain_code, sequence_code, lookup):
    residue_name = get_residue_name_from_lookup(chain_code, sequence_code, lookup)
    if residue_name == UNUSED:
        msg = f"""
                            There is no residue with chain {chain_code} and sequence code {sequence_code} in the input
                            sequence so I can't find a residue name for this residue
                        """
        raise BadSequenceException(msg)

    return residue_name


def _tabulate_sequence(entry: Entry) -> None:
    sequence = sequence_from_entry(entry)
    table = []
    for residue in sequence:
        row = [residue.chain_code, residue.sequence_code, residue.residue_name]
        table.append(row)

    if not table:
        result = "NO SEQUENCE..!"
    else:
        headers = "chain_code sequence_code residue_name".split()
        headers = [header.replace("_", " ") for header in headers]
        result = tabulate(table, headers=headers)

    return result



from pathlib import Path
from typing import List

import typer
from fyeah import f
from pynmrstar import Entry, Loop

from nef_pipelines.lib.nef_lib import (
    NEF_PIPELINES_PREFIX,
    UNUSED,
    create_nef_save_frame,
    read_file_or_exit,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.sequence_lib import (
    exit_if_chain_not_in_entrys_sequence,
    sequence_from_entry,
)
from nef_pipelines.transcoders.nmrpipe.nmrpipe_lib import read_db_file_records
from nef_pipelines.transcoders.talos import import_app
from nef_pipelines.transcoders.talos.talos_lib import read_order_parmeters

app = typer.Typer()

FRAME_NAME_HELP = """\
                        name for the frame that will be created,
                        default is {chain_code}_{element_1}{isotope_number_1}_{element_2}{isotope_number_2}
                    """


PIPELINES_ORDER_DATA_CATEGORY = f"{NEF_PIPELINES_PREFIX}_order_data"
DEFAULT_SEPARATOR = "::"
DEFAULT_END = ""


@import_app.command()
def order_parameters(
    chain_code: str = typer.Option(
        "A",
        "--chain",
        help="chain codes for the new chain",
        metavar="<CHAIN-CODE>",
    ),
    frame_name: str = typer.Option(
        "{chain_code}_{element_1}{isotope_number_1}_{element_2}{isotope_number_2}",
        "-f",
        "--frame",
        help=FRAME_NAME_HELP,
    ),
    input: Path = typer.Option(
        None,
        "-i",
        "--input",
        metavar="|INPUT|",
        help="file to read NEF data from",
    ),
    file_name: Path = typer.Argument(
        ..., help="input files of type <TALOS-FILE>.pred", metavar="<PRED-FILE>"
    ),
):
    """-  convert rci order parameters (S2) from a talos file to NEF"""

    entry = read_or_create_entry_exit_error_on_bad_file(input)

    lines = read_file_or_exit(file_name)

    entry = pipe(entry, lines, chain_code, frame_name)

    print(entry)


def pipe(entry: Entry, lines: List[str], chain_code: str, frame_name: str):
    exit_if_chain_not_in_entrys_sequence(chain_code, entry)

    nef_sequence = sequence_from_entry(entry)

    gdb_records = read_db_file_records(lines)

    talos_order = read_order_parmeters(
        gdb_records,
        chain_code=chain_code,
        file_name=entry.source,
        nef_sequence=nef_sequence,
    )

    order_data_by_dipoles = {}
    for value in talos_order.values:
        atom = value.atom
        dipole_atom = value.dipole_atom
        dipole_pair = (atom.element, atom.isotope_number), (
            dipole_atom.element,
            dipole_atom.isotope_number,
        )
        atom_pair_values = order_data_by_dipoles.setdefault(dipole_pair, {})

        atom_pair = atom, dipole_atom

        value_set = atom_pair_values.setdefault(atom_pair, {})

        value_set[value.value_type] = (value.value, value.value_error)

    for isotopes, values in order_data_by_dipoles.items():

        # used in f function below
        (element_1, isotope_number_1), (element_2, isotope_number_2) = isotopes
        frame_name = f(frame_name)
        frame_id = f(frame_name)

        frame = create_nef_save_frame(PIPELINES_ORDER_DATA_CATEGORY, frame_id)

        extra_tags = {
            "data_type": "model_free",
            "relaxation_atom_id": "2",
            "source": "estimate",
            "diffusion_model": UNUSED,
        }

        for tag, value in extra_tags.items():
            frame.add_tag(tag, value)

        data_loop = Loop.from_scratch(f"{NEF_PIPELINES_PREFIX}_order_values")

        value_types = set()
        for value in values.values():
            value_types.update(value.keys())

        type_strings = [f"{str(type)}  {str(type)}_err" for type in value_types]
        type_string = " ".join(type_strings)

        tags = f"""
            index
            chain_code_1 sequence_code_1 residue_name_1 atom_name_1
            chain_code_2 sequence_code_2 residue_name_2 atom_name_2
            {type_string}
            """.lower().split()

        for tag in tags:
            data_loop.add_tag(tag)

        frame.add_loop(data_loop)

        for index, ((atom_1, atom_2), value_dict) in enumerate(values.items(), start=1):
            residue_1 = atom_1.residue
            residue_2 = atom_2.residue
            row = {
                "index": index,
                "chain_code_1": residue_1.chain_code,
                "sequence_code_1": residue_1.sequence_code,
                "residue_name_1": residue_1.residue_name,
                "atom_name_1": atom_1.atom_name,
                "chain_code_2": residue_2.chain_code,
                "sequence_code_2": residue_2.sequence_code,
                "residue_name_2": residue_2.residue_name,
                "atom_name_2": atom_2.atom_name,
            }

            for value_type, (value, error) in value_dict.items():
                value_type = value_type.lower()
                row[str(value_type)] = value
                row[f"{value_type}_err"] = error

            data_loop.add_data(
                [
                    row,
                ]
            )

        entry.add_saveframe(frame)

        return entry



import string
from enum import auto

from strenum import UppercaseStrEnum

# from https://www.kherb.io/docs/nmr_table.html
# The values are taken from the IAEA's table of recommended nuclear magnetic moments and have been converted to
# gyromagnetic ratios using the 2018 CODATA recommended values for the nuclear magneton and Planck's constant.
# However, the value for hydrogen has been replaced with the CODATA value. The number in parentheses represents
# the one-sigma (1σ) uncertainty in the last two digits of the value provided.
# this is agreement with the values in the Cavanagh et als book (2nd edition)
# 42.577478518 * 2 * 3.1415927 / 100 = 2.675221874410521 vs 2.6752 in Cavanagh within reasonable errors
# the value from NIST is 2.675 221 8708 (12) s-1 T-1 which is in agreement with the value in Cavanagh
MAGNETOGYRIC_RATIO_1H = 42.577478518  # uncertainty (18)  MHz/T


class Isotope(UppercaseStrEnum):
    H1 = auto()
    H2 = auto()
    H3 = auto()
    N15 = auto()
    C13 = auto()
    O17 = auto()
    F19 = auto()
    P31 = auto()

    def __str__(self):
        element = self.name.rstrip(string.digits)
        isotope_number = self.name[len(element) :]
        return f"{isotope_number}{element}"


# fmt: off
GAMMA_RATIOS = {
    # where did i get these from (not wikipedia or the bruker handbook)? is that an authoritative source?
    # Nuc	        ratio	        # (MHz⋅T−1)
    Isotope.H1:     1.0,            # 42.58
    Isotope.H2:     0.153508386,    # 6.54
    Isotope.H3:     1.066643718,    # 45.42
    Isotope.C13:    0.251503855,    # 10.71
    Isotope.N15:    0.101368145,    # -4.32 - should really be negative !
    Isotope.O17:    0.135564627,    # -5.77 - should really be negative !
    Isotope.F19:    0.94129576,     # 40.08
    Isotope.P31:    0.404791467,    # 17.24
}

ATOM_TO_ISOTOPE = {
    # Nuc   Isotope
    "H": Isotope.H1,
    "D": Isotope.H2,
    "T": Isotope.H3,
    "C": Isotope.C13,
    "N": Isotope.N15,
    "O": Isotope.O17,
    "F": Isotope.F19,
    "P": Isotope.P31,
}

CODE_TO_ISOTOPE = {
    # Nuc   Isotope
    "1H": Isotope.H1,
    "2H": Isotope.H2,
    "3H": Isotope.H3,
    "13C": Isotope.C13,
    "15N": Isotope.N15,
    "17O": Isotope.O17,
    "19F": Isotope.F19,
    "31P": Isotope.P31,
}
# fmt: on



import typer

from nef_pipelines import nef_app

shifts_app = typer.Typer()


if nef_app.app:
    nef_app.app.add_typer(
        shifts_app, name="shifts", help="- carry out operations on shifts [average]"
    )

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.tools.shifts.average  # noqa: F401



import typer

from nef_pipelines import nef_app

app = typer.Typer()
import_app = typer.Typer()
export_app = typer.Typer()


if nef_app.app:

    nef_app.app.add_typer(
        app,
        name="talos",
        help="- read and write talos files [shifts & restraints]",
    )

    app.add_typer(import_app, name="import", help="- import talos dihderal restraints ")
    app.add_typer(
        export_app, name="export", help="-  export talos [shifts (& sequence)]"
    )

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.transcoders.talos.exporters.shifts  # noqa: F401
    import nef_pipelines.transcoders.talos.importers.order_parameters  # noqa: F401
    import nef_pipelines.transcoders.talos.importers.restraints  # noqa: F401
    import nef_pipelines.transcoders.talos.importers.secondary_structure  # noqa: F401
    import nef_pipelines.transcoders.talos.importers.sequence  # noqa: F401



import string
import sys

import requests
from fyeah import f
from pynmrstar import Entry
from requests.exceptions import HTTPError
from tabulate import tabulate

from nef_pipelines.lib.util import exit_error, is_int
from nef_pipelines.transcoders.nmrstar.importers.project_shortcuts import (
    SHORTCUT_URLS,
    SHORTCUTS,
)
from nef_pipelines.transcoders.nmrstar.importers.sequence import pipe as sequence_pipe
from nef_pipelines.transcoders.nmrstar.importers.shifts import pipe as shift_pipe


def pipe(
    nef_entry,
    nmrstar_entry,
    chain_codes,
    no_chain_starts,
    no_chain_ends,
    use_author,
    stereo_mode,
    file_path,
):

    entry_id = nef_entry.entry_id
    nef_entry = sequence_pipe(
        nef_entry,
        chain_codes,
        [],
        no_chain_starts,
        no_chain_ends,
        entry_id,
        nmrstar_entry,
        file_path,
        use_author,
    )

    # def pipe(
    #         nef_entry: Entry,
    #         chain_codes: List[str],
    #         frame_name: str,
    #         nmrstar_entry: Entry,
    #         file_name: Path,
    #         use_author: bool,
    #         stereo_mode: StereoAssignmentHandling,
    # ):

    nef_entry = shift_pipe(
        nef_entry, [], entry_id, nmrstar_entry, file_path, use_author, stereo_mode
    )

    return nef_entry


def _list_shortcuts_and_exit():
    print("The available shortcuts are:\n", file=sys.stderr)
    shortcuts = [
        [shortcut.lower(), entry, SHORTCUT_URLS[shortcut]]
        for shortcut, entry in SHORTCUTS.items()
    ]
    HEADERS = ["name", "entry id.", "url"]
    print(tabulate(shortcuts, headers=HEADERS), file=sys.stderr)
    print("\nexiting...", file=sys.stderr)
    sys.exit(0)


def _parse_text_to_star_or_none(possible_entry):
    nmrstar_entry = None
    if possible_entry and possible_entry.startswith(b"data_"):
        try:
            nmrstar_entry = Entry.from_string(possible_entry.decode("utf-8"))
        except Exception:
            pass
    return nmrstar_entry


def _get_bmrb_entry_from_web_or_none(url, exit_on_error=True):
    possible_entry = None
    error = False
    exception = None
    if url:
        try:
            response = requests.get(url)
            response.raise_for_status()
            possible_entry = response.content
        except HTTPError as http_err:
            msg = f"while trying to download the url {url} there was an http error {http_err}"
            exception = http_err
        except Exception as err:
            msg = f"while trying to download the url {url} there was an error {err}"
            exception = err

    if error:
        if exit_on_error:
            exit_error(msg, exception)
        else:
            print(f"WARNING: {msg}, trying to read as a local file", file=sys.stderr)

    return possible_entry


def _get_path_as_url_or_none(file_path, url_template):
    url = None
    is_bmrb = False
    if file_path.startswith("https://"):
        url = file_path
    elif file_path.startswith("bmr"):
        entry_number = file_path[3:]
        entry_check = entry_number.lstrip(string.digits)
        if len(entry_check) == 0:
            url = f(url_template)
            is_bmrb = True
    elif is_int(file_path):
        entry_number = file_path
        url = f(url_template)
        is_bmrb = True

    return url, is_bmrb



from argparse import Namespace
from pathlib import Path
from typing import List

import typer

from nef_pipelines.lib.sequence_lib import get_chain_code_iter, sequence_to_nef_frame
from nef_pipelines.lib.typer_utils import get_args
from nef_pipelines.lib.util import (
    STDIN,
    exit_error,
    parse_comma_separated_options,
    process_stream_and_add_frames,
)
from nef_pipelines.transcoders.nmrview import import_app
from nef_pipelines.transcoders.nmrview.nmrview_lib import read_sequence

app = typer.Typer()


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def sequence(
    chain_codes: str = typer.Option(
        "A",
        "--chains",
        help="chain codes, can be called multiple times and or be a comma separated list [no spaces!]",
        metavar="<CHAIN-CODES>",
    ),
    no_chain_start: bool = typer.Option(
        False,
        "--no-chain-start/",
        help="don't include a start of chain link type for the first residue",
    ),
    no_chain_end: bool = typer.Option(
        False,
        "--no-chain-end/",
        help="don't include an end of chain link type for the last residue",
    ),
    entry_name: str = typer.Option("nmrview", help="a name for the entry"),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        help="pipe to read NEF data from other than stdin",
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="input files of type nmrview.seq", metavar="<SEQ-FILE>"
    ),
):
    """convert nmrview sequence file <nmrview>.seq files to NEF"""
    try:
        args = get_args()

        process_sequence(args)
    except Exception as e:
        exit_error(f"reading sequence failed because {e}")


def process_sequence(args: Namespace):
    nmrview_frames = []

    chain_codes = parse_comma_separated_options(args.chain_codes)
    chain_code_iter = get_chain_code_iter(chain_codes)
    for file_name, chain_code in zip(args.file_names, chain_code_iter):
        with open(file_name, "r") as lines:
            nmrview_sequence = read_sequence(lines, chain_code=chain_code)

            frame = sequence_to_nef_frame(nmrview_sequence, args.entry_name)

            nmrview_frames.append(frame)

    entry = process_stream_and_add_frames(nmrview_frames, args)

    print(entry)


if __name__ == "__main__":

    typer.run(sequence)



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    read_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.xeasy.importers.shifts import shifts

SHIFTS_XEASY = "nef_chemical_shift_list_xeasy"

app = typer.Typer()
app.command()(shifts)

EXPECTED = """\
save_nef_chemical_shift_list_xeasy
   _nef_chemical_shift_list.sf_category   nef_chemical_shift_list
   _nef_chemical_shift_list.sf_framecode  nef_chemical_shift_list_xeasy
   loop_
      _nef_chemical_shift.chain_code
      _nef_chemical_shift.sequence_code
      _nef_chemical_shift.residue_name
      _nef_chemical_shift.atom_name
      _nef_chemical_shift.value
      _nef_chemical_shift.value_uncertainty
      _nef_chemical_shift.element
      _nef_chemical_shift.isotope_number
     A   1   HIS   N     112.794   0.010   .   .
     A   1   HIS   HE1   7.955     0.010   .   .
     A   2   MET   N     113.996   0.000   .   .
     A   2   MET   QE    3.778     0.010   .   .
     A   2   MET   CE    27.054    0.010   .   .
     A   3   ARG   HD2   3.133     0.000   .   .
     A   3   ARG   HD3   3.134     0.000   .   .
     A   3   ARG   HE    4.378     0.010   .   .
   stop_
save_

"""


def test_ppm_out_short():

    STREAM = read_test_data("basic_sequence.nef", __file__)

    path = path_in_test_data(__file__, "basic_shifts.prot")

    result = run_and_report(app, [path], input=STREAM)

    shifts_result = isolate_frame(result.stdout, SHIFTS_XEASY)

    assert_lines_match(EXPECTED, shifts_result)



import string
import sys
from collections import Counter
from enum import IntEnum, auto
from pathlib import Path
from typing import Dict, List, Tuple, Union

import typer
from ordered_set import OrderedSet
from parse import compile as parse_compile
from pynmrstar import Entry, Loop, Saveframe
from strenum import StrEnum
from tabulate import tabulate

from nef_pipelines.lib.nef_lib import (
    NEF_PIPELINES_PREFIX,
    NEF_RELAXATION_VERSION,
    UNUSED,
    SelectionType,
    add_frames_to_entry,
    create_nef_save_frame,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames,
)
from nef_pipelines.lib.util import (
    exit_error,
    is_float,
    is_int,
    parse_comma_separated_options,
    strings_to_tabulated_terminal_sensitive,
)
from nef_pipelines.tools.series import series_app

NAMESPACE = NEF_PIPELINES_PREFIX

FRAMES_AND_TIMES_HELP = "the list of frame selectors and their values"


class RelaxationExperimentType(StrEnum):
    AUTO_RELAXATION = auto()
    DIPOLE_CSA_CROSS_CORRELATIONS = auto()
    DIPOLE_DIPOLE_CROSS_CORRELATIONS = auto()
    DIPOLE_DIPOLE_RELAXATION = auto()
    HETERONUCLEAR_NOES = auto()
    HETERONUCLEAR_R1_RELAXATION = auto()
    HETERONUCLEAR_R1RHO_RELAXATION = auto()
    HETERONUCLEAR_R2_RELAXATION = auto()
    H_EXCHANGE_PROTECTION_FACTORS = auto()
    H_EXCHANGE_RATES = auto()
    HOMONUCLEAR_NOES = auto()
    CPMG = auto()
    CEST = auto()
    OTHER = auto()
    T1 = HETERONUCLEAR_R1_RELAXATION
    T2 = HETERONUCLEAR_R2_RELAXATION
    T1RHO = HETERONUCLEAR_R1RHO_RELAXATION
    R1 = HETERONUCLEAR_R1_RELAXATION
    R2 = HETERONUCLEAR_R2_RELAXATION
    NOE = HETERONUCLEAR_NOES


EXPERIMENT_TYPE_TO_SERIES_VARIABLE_TYPE = {
    RelaxationExperimentType.HETERONUCLEAR_R1_RELAXATION: "time",
    RelaxationExperimentType.HETERONUCLEAR_R2_RELAXATION: "time",
    RelaxationExperimentType.HETERONUCLEAR_R1RHO_RELAXATION: "time",
    RelaxationExperimentType.HETERONUCLEAR_R1_RELAXATION: "time",
    RelaxationExperimentType.HETERONUCLEAR_R2_RELAXATION: "time",
    RelaxationExperimentType.HETERONUCLEAR_NOES: "saturation",
    RelaxationExperimentType.CPMG: "cycles",
    RelaxationExperimentType.CEST: "offset",
    RelaxationExperimentType.OTHER: "unknown",
}

EXPERIMENT_TYPE_TO_DATA_VARIABLE_TYPE = {
    RelaxationExperimentType.HETERONUCLEAR_R1_RELAXATION: "intensity",
    RelaxationExperimentType.HETERONUCLEAR_R2_RELAXATION: "intensity",
    RelaxationExperimentType.HETERONUCLEAR_R1RHO_RELAXATION: "intensity",
    RelaxationExperimentType.HETERONUCLEAR_R1_RELAXATION: "intensity",
    RelaxationExperimentType.HETERONUCLEAR_R2_RELAXATION: "intensity",
    RelaxationExperimentType.HETERONUCLEAR_NOES: "intensity",
    RelaxationExperimentType.CPMG: "intensity",
    RelaxationExperimentType.CEST: "intensity",
    RelaxationExperimentType.OTHER: "unknown",
}

DATA_VARIABLE_TYPE_TO_DATA_VARIABLE_UNIT = {
    "intensity": UNUSED,
    "volume": UNUSED,
    "unknown": UNUSED,
}

SERIES_VARIABLE_TYPE_TO_SERIES_VARIABLE_UNIT = {
    "time": "s",
    "saturation": "fraction",
    "cycles": UNUSED,
    "offset": "hz",
}

EXPERIMENT_TO_IDENTIFIER = {
    RelaxationExperimentType.AUTO_RELAXATION: "auto_relaxation",
    RelaxationExperimentType.DIPOLE_CSA_CROSS_CORRELATIONS: "dipole_CSA_cross_correlations",
    RelaxationExperimentType.DIPOLE_DIPOLE_CROSS_CORRELATIONS: "dipole_dipole_cross_correlations",
    RelaxationExperimentType.DIPOLE_DIPOLE_RELAXATION: "dipole_dipole_relaxation",
    RelaxationExperimentType.HETERONUCLEAR_NOES: "heteronuclear_NOEs",
    RelaxationExperimentType.HETERONUCLEAR_R1_RELAXATION: "heteronuclear_R1_relaxation",
    RelaxationExperimentType.HETERONUCLEAR_R1RHO_RELAXATION: "heteronuclear_R1rho_relaxation",
    RelaxationExperimentType.HETERONUCLEAR_R2_RELAXATION: "heteronuclear_R2_relaxation",
    RelaxationExperimentType.H_EXCHANGE_PROTECTION_FACTORS: "H_exchange_protection_factors",
    RelaxationExperimentType.H_EXCHANGE_RATES: "H_exchange_rates",
    RelaxationExperimentType.HOMONUCLEAR_NOES: "homonuclear_NOEs",
    RelaxationExperimentType.CPMG: "CPMG",
    RelaxationExperimentType.CEST: "CEST",
    RelaxationExperimentType.OTHER: "other",
}

NAME_TO_EXPERIMENT_TYPE = {
    "T1": RelaxationExperimentType.HETERONUCLEAR_R1_RELAXATION,
    "T2": RelaxationExperimentType.HETERONUCLEAR_R2_RELAXATION,
    "T1rho": RelaxationExperimentType.HETERONUCLEAR_R1RHO_RELAXATION,
    "T1p": RelaxationExperimentType.HETERONUCLEAR_R1RHO_RELAXATION,
    "R1": RelaxationExperimentType.HETERONUCLEAR_R1_RELAXATION,
    "R2": RelaxationExperimentType.HETERONUCLEAR_R2_RELAXATION,
    "NOE": RelaxationExperimentType.HETERONUCLEAR_NOES,
    "CPMG": RelaxationExperimentType.CPMG,
    "REX": RelaxationExperimentType.CPMG,
    "CEST": RelaxationExperimentType.CEST,
    "DEST": RelaxationExperimentType.CEST,
}


class SaturationState(IntEnum):
    SATURATED = 1
    UNSATURATED = 0


STR_TO_SATURATION_STATE = {
    "saturated": SaturationState.SATURATED,
    "unsaturated": SaturationState.UNSATURATED,
    "on": SaturationState.SATURATED,
    "off": SaturationState.UNSATURATED,
    "true": SaturationState.SATURATED,
    "false": SaturationState.UNSATURATED,
    "yes": SaturationState.SATURATED,
    "no": SaturationState.UNSATURATED,
    "1": SaturationState.SATURATED,
    "0": SaturationState.UNSATURATED,
}


@series_app.command()
def build(
    input: Path = typer.Option(
        None,
        "-i",
        "--in",
        metavar="NEF-FILE",
        help="read NEF data from a file instead of stdin",
    ),
    timings: List[str] = typer.Option(
        None,
        help="the list of relaxation times or values",
    ),
    input_unit: str = typer.Option(
        None,
        "--unit",
        help="the unit of the relaxation times",
    ),
    name: str = typer.Option(
        None,
        help="the name of the series, if not provided it will be guessed from the frames",
    ),
    experiment_type: str = typer.Option(None, help="the type of the experiment"),
    display_parsed_values: bool = typer.Option(False, help="display the parsed values"),
    frames_selectors: List[str] = typer.Argument(
        None,
        help=FRAMES_AND_TIMES_HELP,
    ),
):
    """- build a NEF data series from a set of spectra [alpha]"""

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    frame_selectors = parse_comma_separated_options(frames_selectors)

    _exit_if_no_frame_selectors(frame_selectors)

    frames = _select_relaxation_frames(entry, frames_selectors)

    frame_timings_and_units = _get_timings_and_units_for_frames(
        frames, frame_selectors, timings, input_unit, display_parsed_values
    )

    frames_and_timings, unit = (
        _ensure_units_consistent_and_get_unit_and_timings_or_exit(
            frame_timings_and_units
        )
    )

    # TODO arenn't these the same? add a name and expt attribute to the template and only use this notherwise
    if not name:
        name = _guess_series_name(entry, frames_and_timings)

    if not experiment_type:
        experiment_type = _guess_experiment_type(name)

    entry = pipe(entry, frames_and_timings, unit, name, experiment_type)

    print(entry)


def pipe(
    entry: Entry,
    frame_timings: Dict[Tuple[str, int], Union[int, float, bool]],
    unit: str,
    name: str,
    experiment_type: str,
) -> Entry:

    series_frame = create_nef_save_frame(f"{NAMESPACE}_series_list", name)

    series_variable_type = EXPERIMENT_TYPE_TO_SERIES_VARIABLE_TYPE[
        RelaxationExperimentType[experiment_type]
    ]
    series_variable_unit = SERIES_VARIABLE_TYPE_TO_SERIES_VARIABLE_UNIT[
        series_variable_type
    ]

    series_variable_unit = (
        unit if unit != series_variable_unit else series_variable_unit
    )

    data_value_type = EXPERIMENT_TYPE_TO_DATA_VARIABLE_TYPE[
        RelaxationExperimentType[experiment_type]
    ]
    data_value_unit = DATA_VARIABLE_TYPE_TO_DATA_VARIABLE_UNIT[data_value_type]

    experiment_type = EXPERIMENT_TO_IDENTIFIER[
        RelaxationExperimentType[experiment_type]
    ]
    tags = [
        ("experiment_type", experiment_type),
        ("series_variable_type", series_variable_type),
        ("series_variable_unit", series_variable_unit),
        ("data_variable_type", series_variable_type),
        ("data_variable_unit", series_variable_unit),
        ("data_value_type", data_value_type),
        ("data_value_unit", data_value_unit),
        ("comment", UNUSED),
    ]
    series_frame.add_tags(tags)

    # TODO: pynmstar doesn't catch prepended spaces here it just gives and obscure error report!
    series_experiment_loop = Loop.from_scratch(f"{NAMESPACE}_series_experiment")
    series_frame.add_loop(series_experiment_loop)
    series_experiment_loop.add_tag(
        [
            "version",
            "nmr_spectrum_id",
            "reference_experiment",
            "combination_id",
            "pseudo_dimension",
            "pseudo_dimension_point",
            "series_variable",
            "series_variable_error",
        ]
    )

    loop_data = []
    for frame_key, value in frame_timings.items():
        frame_name, _ = frame_key
        loop_data.append(
            {
                "version": NEF_RELAXATION_VERSION,
                "nmr_spectrum_id": frame_name,
                "reference_experiment": "false",
                "combination_id": UNUSED,
                "pseudo_dimension": UNUSED,
                "pseudo_dimension_point": UNUSED,
                "series_variable": value,
                "series_variable_error": UNUSED,
            }
        )

    series_experiment_loop.add_data(loop_data)

    add_frames_to_entry(
        entry,
        [
            series_frame,
        ],
    )

    return entry


def _guess_experiment_type(name):
    upper_name = name.upper()
    for experiment_name in NAME_TO_EXPERIMENT_TYPE:
        if experiment_name in upper_name:
            result = NAME_TO_EXPERIMENT_TYPE[upper_name]
            break
        else:
            result = "unknown"
    return result


def _guess_series_name(entry, frames_and_timings):
    result = None
    for relaxation_type in NAME_TO_EXPERIMENT_TYPE:
        do_break = False
        for name, _ in frames_and_timings.keys():
            name = name.upper()
            if relaxation_type in name.upper():
                start_index = name.index(relaxation_type)
                end_index = start_index + len(relaxation_type)
                found_relaxation_type = name[start_index:end_index]

                result = found_relaxation_type
                do_break = True
                break
        if do_break:
            break

    if not result:
        result = "unknown_type"

    bad_frame_name = False
    for frame in entry:
        if frame.name == result:
            bad_frame_name = True
            break

    if bad_frame_name:
        base_result = result
        for i in range(1, 100):
            result = f"series`{i}`"
            bad_frame_name = False
            for frame in entry:
                if frame.name == result:
                    bad_frame_name = True
                    break

    if bad_frame_name:
        _exit_if_too_many_clashing_frame_names(bad_frame_name, base_result)

    return result


def _exit_if_too_many_clashing_frame_names(bad_frame_name, base_result):
    if bad_frame_name:
        msg = f"""
            after searching the series of names {base_result}`1` to {base_result}`100` i could not find a name
            that was not already in the entry, choose a name on the command line, you have way too much data!
            more serious report this as a bug...
        """
        exit_error(msg)


def _get_timings_and_units_for_frames(
    frames, frame_selectors, timings, input_unit, display_parsed_values
):

    if timings:
        timings_and_units = _parse_timings(timings)
        units = [timing_and_unit[1] for timing_and_unit in timings_and_units]

        _exit_if_input_unit_defined_and_any_unit_parsed(
            input_unit, units, timings_and_units
        )

        if input_unit:
            timings_and_units = [
                (timing, input_unit) for timing, _ in timings_and_units
            ]

        _exit_error_it_number_of_timings_doesnt_match_number_of_frames(
            frames, timings_and_units
        )

        frame_timings_and_units = {
            (frame.name, id(frame)): timing_and_unit
            for frame, timing_and_unit in zip(frames, timings_and_units)
        }
    else:
        _exit_if_frame_selector_has_no_timing_selector(frame_selectors)
        frame_timings_and_units = _parse_timings_from_frames_or_exit(
            frames, frame_selectors[0], display_parsed_values
        )

    return frame_timings_and_units


def _exit_if_input_unit_defined_and_any_unit_parsed(
    input_unit, units, timings_and_units
):
    if input_unit and any(unit != "" for unit in units):
        units_and_timings_strings = [
            f"{unit}{timing}" for unit, timing in timings_and_units if unit
        ]
        msg = f"""
                you cannot provide a unit on the command line [--unit] and units in the timings [--timings]
                at the same time command line unit was {input_unit} and the units and timings were
                {strings_to_tabulated_terminal_sensitive(units_and_timings_strings)}

                """
        exit_error(msg)


def _exit_error_it_number_of_timings_doesnt_match_number_of_frames(
    frames, timings_and_units
):
    if len(timings_and_units) != len(frames):
        msg = f"the number of timings {len(timings_and_units)} does not match the number of frames {len(frames)}"
        exit_error(msg)


def _exit_if_frame_selector_has_no_timing_selector(frame_selectors):
    if len(frame_selectors) != 1 or "{var}" not in frame_selectors[0]:
        NEWLINE = "\n"
        msg = f"""
                    no timings provided to select timings you must have a single frame selector with
                    a {{var}} placeholder i got {len(frame_selectors)} frame selectors and they were:
                    {NEWLINE.join(frame_selectors)}
                """
        exit_error(msg)


def _select_relaxation_frames(entry: Entry, frame_selectors: List[str]) -> List[str]:
    spectrum_frames = select_frames(entry, "nef_nmr_spectrum", SelectionType.CATEGORY)
    spectrum_frame_ids = OrderedSet(
        [
            (spectrum_frame.name, id(spectrum_frame))
            for spectrum_frame in spectrum_frames
        ]
    )

    frame_selectors = [selector.replace("{var}", "*") for selector in frame_selectors]
    frame_selectors = [selector.replace("{}", "*") for selector in frame_selectors]

    relaxation_frames = []
    for frame_selector in frame_selectors:
        relaxation_frames.extend(
            select_frames(entry, frame_selector, SelectionType.NAME)
        )

    relaxation_frame_ids = OrderedSet(
        [
            (relaxation_frame.name, id(relaxation_frame))
            for relaxation_frame in relaxation_frames
        ]
    )

    # TODO: just do an intersection dummy!
    if not relaxation_frame_ids.issubset(spectrum_frame_ids):
        non_spectrum_frames = relaxation_frame_ids.difference(spectrum_frame_ids)
        non_spectrum_frame_names = "\n".join(
            [non_spectrum_frame[0] for non_spectrum_frame in non_spectrum_frames]
        )
        # TODO: just filter out non spectra frames...
        _exit_if_some_frames_are_not_spectra(non_spectrum_frame_names)

    return relaxation_frames


def _exit_if_some_frames_are_not_spectra(non_spectrum_frame_names):
    msg = f"""
            some of the selected frames are not spectrum frames:
            {non_spectrum_frame_names}
            """
    exit_error(msg)


def _parse_timings_from_frames_or_exit(
    save_frames: List[Saveframe], frame_selector: str, display_parsed_values: bool
) -> Dict[Tuple[str, int], List[Tuple[Union[int, float, bool], str]]]:
    if "{var}" not in frame_selector:
        msg = f"no timing selector placeholder {{var}} provided in frame selector {frame_selector}"
        exit_error(msg)

    frame_selector = frame_selector.replace("{var}", "{var}")
    frame_selector = frame_selector.replace("*", "{:opt}")

    parser = parse_compile(frame_selector)

    parsed_values = []

    frame_value_and_units = {}
    for frame in save_frames:
        frame_key = frame.name, id(frame)
        result = parser.search(frame.name)

        value, unit = None, None
        if result and "var" in result:
            raw_value = result["var"]
            raw_value = raw_value.replace("_", ".")
            raw_value = raw_value.replace("-", ".")

            value, unit = _parse_value_and_unit(raw_value)

        if display_parsed_values:
            parsed_values.append([frame.name, raw_value, value, unit])

        if value is None:
            _exit_if_no_variable_found(frame, frame_selector)

        if not (is_float(value) or value.lower() in "on off true false yes no"):
            msg = f"the variable {value} in frame {frame.name} is not a number or a boolean"
            exit_error(msg)

        frame_value_and_units[frame_key] = value, unit

    if display_parsed_values:
        print(
            tabulate(parsed_values, headers=["spectrum", "raw value", "value", "unit"]),
            file=sys.stderr,
        )
        sys.exit(0)

    return frame_value_and_units


def _exit_if_no_variable_found(frame, frame_selector):
    msg = f"no variable found in frame {frame.name} with the selector {frame_selector}"
    exit_error(msg)


def _parse_timings(timings: List[str]) -> List[Tuple[Union[int, float, bool], str]]:
    timings = parse_comma_separated_options(timings)
    parsed_timings = []
    for timing in timings:
        value, unit = _parse_value_and_unit(timing)
        parsed_timings.append((value, unit))
    return parsed_timings


def _parse_value_and_unit(value: str) -> Tuple[Union[bool, int, float, None]]:

    unit = value.lstrip(string.digits + ".eE")
    len_digits = len(value) - len(unit)
    value = value[:len_digits]

    lower_unit = unit.lower()
    if value and unit:
        pass
    elif not value and lower_unit in STR_TO_SATURATION_STATE:
        unit = "bool"
        value = STR_TO_SATURATION_STATE[lower_unit]
    elif not unit:
        if is_int(value):
            value = int(value)
        elif is_float(value):
            value = float(value)
    else:
        unit = None
        value = None

    return value, unit


def _exit_if_no_frame_selectors(frame_selectors):
    if len(frame_selectors) == 0:
        msg = "you must select some frames!"
        exit_error(msg)


def _ensure_units_consistent_and_get_unit_and_timings_or_exit(frames_timings_and_units):

    _exit_units_not_consistent(frames_timings_and_units)

    _exit_if_numbers_and_bools_are_mixed(frames_timings_and_units)

    float_counter = Counter()
    int_counter = Counter()
    for frame, (value, unit) in frames_timings_and_units.items():
        if isinstance(value, int):
            int_counter[value] += 1
        elif isinstance(value, float):
            float_counter[value] += 1

    if int_counter and float_counter:
        for frame_key, (value, unit) in frames_timings_and_units.items():
            value = float(value)

            frames_timings_and_units[frame_key] = value, unit

    unit = list(frames_timings_and_units.values())[0][1]
    frames_and_timings = {
        frame_key: value for frame_key, (value, _) in frames_timings_and_units.items()
    }

    return frames_and_timings, unit


def _exit_if_numbers_and_bools_are_mixed(frame_timings_and_units):

    counter_numbers = Counter()
    couter_booleans = Counter()

    for frame, (value, unit) in frame_timings_and_units.items():
        if isinstance(value, bool):
            couter_booleans[value] += 1
        elif is_float(value):
            counter_numbers[value] += 1

    if counter_numbers and couter_booleans:
        numbers_strings = ", ".join(counter_numbers.keys())
        booleans_strings = ", ".join(couter_booleans.keys())
        msg = f"""
                you cannot have both numbers and booleans in the values
                the booleans were
                {booleans_strings}
                the numbers were
                {numbers_strings}
            """
        exit_error(msg)


def _exit_units_not_consistent(frame_timings_and_units):

    unit_counter = Counter()
    for frame, (value, unit) in frame_timings_and_units.items():
        unit_counter[unit] += 1

    if len(unit_counter) > 1:
        unit_strings = ", ".join(unit_counter.keys())
        msg = f"""
            the units of the timings are not consistent, the units found were
            {unit_strings}
            """
        exit_error(msg)


def _exit_if_units_are_not_compatible(frame_timings_and_units):
    units = {unit for timing, unit in frame_timings_and_units.values()}
    if len(units) > 1:
        units_and_timings_strings = [
            f"{unit}{timing}" for timing, unit in frame_timings_and_units.values()
        ]
        msg = f"""
                the units of the timings are not compatible they are {units}
                the timings and units were
                {strings_to_tabulated_terminal_sensitive(units_and_timings_strings)}
                """
        exit_error(msg)



from copy import copy
from itertools import tee
from pathlib import Path
from textwrap import dedent
from typing import Dict, List, Union

import typer
from pynmrstar import Entry
from tabulate import tabulate
from typer import Argument, Option

from nef_pipelines.lib.nef_lib import (
    NEF_MOLECULAR_SYSTEM,
    SELECTORS_LOWER,
    SelectionType,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames,
)
from nef_pipelines.lib.sequence_lib import (
    chains_from_frames,
    get_chain_starts,
    select_best_residues_by_info_content,
    sequences_from_frames,
)
from nef_pipelines.lib.util import (
    chunks,
    end_with_ordinal,
    exit_error,
    is_int,
    strings_to_tabulated_terminal_sensitive,
)
from nef_pipelines.tools.chains import chains_app

REFERENCE_FRAMES_HELP = """
frames to find the starting residue of a chain for renumbering when using starts  you can have multiple reference
frames and the lowest residue number found amongst them is used. The option can be repeated  or you can use comma
separated lists of frame names [default: the molecular system frame if present otherwise all frames].
"""

app = typer.Typer()

CHAIN_CODE = "chain_code"

SEQUENCE_CODE = "sequence_code"


# noinspection PyUnusedLocal
@chains_app.command()
def renumber(
    chain_offsets_or_starts: List[str] = Argument(
        None,
        metavar="<CHAIN-CODE> <OFFSET/START>",
        help="chain-codes and offsets/starts for renumbering residue numbers, offsets are used by default and chain "
        "starts are available using the --starts option [default: no selections, no offets; do nothing...]",
        show_default=False,
    ),
    input: Path = typer.Option(
        Path("-"),
        "-i",
        "--in",
        metavar="NEF-FILE",
        help="where to read NEF data from either a file or stdin '-'",
    ),
    selector_type: SelectionType = typer.Option(
        SelectionType.ANY,
        "-s",
        "--selector",
        help=f"control how to select frames to renumber, can be one of: {SELECTORS_LOWER}. "
        "Any will match on names first and then if there is no match attempt to match on category",
    ),
    reference_frame_selectors: List[str] = Option(
        None, "--references", help=REFERENCE_FRAMES_HELP, show_default=False
    ),
    starts: bool = Option(
        False,
        "-s",
        "--starts",
        help="renumber residues using a starting value for the chain "
        "rather than an offset",
    ),
    frame_selectors: List[str] = Option(
        None,
        "-f",
        "--frames",
        metavar="<FRAME-SELECTOR>",
        help="Limit changes to a particular frame by name or category [the selector], note: wildcards are "
        "allowed. Frames are selected by name and then category if the name doesn't match"
        "note: the option -t /--selector-type allows you to choose which selection type to use. "
        "[default: all frames]",
        show_default=False,
    ),
):
    """- renumber chains in a nef file"""

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    if not reference_frame_selectors:
        reference_frames = []

    chain_offsets_or_starts = _get_chain_offset_pairs_or_exit_error(
        chain_offsets_or_starts
    )

    reference_frames = _select_reference_frames(entry, reference_frame_selectors)

    _exit_error_if_no_reference_frames(
        entry, reference_frame_selectors, reference_frames
    )

    if starts:
        reference_chains_residues = sequences_from_frames(reference_frames)
        reference_chain_starts = get_chain_starts(reference_chains_residues)
        chain_offsets = _chain_starts_to_offsets(
            reference_chain_starts, chain_offsets_or_starts
        )
    else:
        chain_offsets = chain_offsets_or_starts

    _exit_if_selected_chain_not_in_frames(
        "reference_chain", reference_frames, input, chain_offsets.keys()
    )
    _exit_error_if_no_chain_offsets(chain_offsets_or_starts, reference_frames, input)

    entry = pipe(entry, frame_selectors, selector_type, chain_offsets)

    print(entry)


def pipe(
    entry: Entry,
    frame_selectors: Union[str, List[str]],
    frame_selector_type: SelectionType,
    offsets: Dict[str, int],
) -> Entry:

    frames = select_frames(entry, frame_selectors, frame_selector_type)

    _exit_if_selected_chain_not_in_frames(
        "chain to renumber", frames, input, offsets.keys()
    )

    offset_chains_in_frames(frames, offsets)

    return entry


def _exit_if_selected_chain_not_in_frames(chain_type, frames, input, chains):
    frame_chains = set()
    for frame in frames:
        frame_chains.update(chains_from_frames(frame))
    frame_chain_names = strings_to_tabulated_terminal_sensitive(frame_chains)

    for chain in chains:

        selected_frames = strings_to_tabulated_terminal_sensitive(
            [frame.name for frame in frames]
        )
        if chain not in frame_chains:
            msg = """
                using the input {input}
                the {chain_type} '{chain}' is not in the frames selected for renumbering

                the selected frames are:

                {selected_frames}

                which have the chains

                {frame_chain_names}
                """
            msg = msg.format(
                chain_type=chain_type,
                input=input,
                chain=chain,
                selected_frames=selected_frames,
                frame_chain_names=frame_chain_names,
            )
            exit_error(msg)


def _exit_error_if_no_chain_offsets(chain_offsets, reference_frames, input):
    if not chain_offsets:
        sequences = [sequences_from_frames(frame) for frame in reference_frames]

        all_residues = set()
        for sequence in sequences:
            all_residues.update(sequence)

        all_residues = select_best_residues_by_info_content(all_residues)

        chain_starts = [
            [chain, start] for chain, start in get_chain_starts(all_residues).items()
        ]

        reference_frame_names = strings_to_tabulated_terminal_sensitive(
            [frame.name for frame in reference_frames]
        )

        if chain_starts:
            headings = "chain start".split()
            table = tabulate(chain_starts, headings)
            table_msg = (
                "here are the starts for the chains in the selected reference frames"
            )
        else:
            table = ""
            table_msg = (
                "also no chain starts were found in the selected reference frames"
            )

        msg = """
                in the file {input}
                you didn't provide any chains and offsets/starts
                {table_msg}

                {table}

                the reference frames for sequences and chain starts were

                {reference_frame_names}
            """
        msg = dedent(msg)
        msg = msg.format(
            input=input,
            table=table,
            table_msg=table_msg,
            reference_frame_names=reference_frame_names,
        )

        exit_error(msg)


def _exit_error_if_no_reference_frames(
    entry, reference_frame_selectors, reference_frames
):
    if not reference_frames:
        msg = """
            no reference frames found using the selection:

            {reference_frame_selectors}

            possible frame names are

            {frame_names}
            """
        msg = dedent(msg)
        reference_frame_selectors = ", ".join(reference_frame_selectors)
        reference_frame_selectors = strings_to_tabulated_terminal_sensitive(
            reference_frame_selectors
        )

        frame_names = strings_to_tabulated_terminal_sensitive(
            [frame.name for frame in entry.frame_list]
        )
        msg = msg.format(
            reference_frame_selectors=reference_frame_selectors, frame_names=frame_names
        )

        exit_error(msg)


def _select_reference_frames(entry, reference_frame_selectors):

    if not reference_frame_selectors:
        reference_frames = select_frames(
            entry, NEF_MOLECULAR_SYSTEM, SelectionType.CATEGORY
        )

        if not reference_frames:
            reference_frames = copy(entry.frame_list)
    else:
        reference_frames = []
        for frame_selector in reference_frame_selectors:
            reference_frames += select_frames(entry, frame_selector, SelectionType.ANY)

    return reference_frames


def _chain_starts_to_offsets(current_chains_starts, new_chain_starts):
    result = {}

    for chain, new_start in new_chain_starts.items():
        old_start = current_chains_starts[chain]

        offset = new_start - old_start
        result[chain] = offset

    return result


def exit_if_molecular_system_isnt_a_singleton(molecular_system_frames):
    number_of_molecular_system_frames = len(molecular_system_frames)
    if number_of_molecular_system_frames == 0:
        exit_error("there must be a molecular system defined to use the starts option")
    elif number_of_molecular_system_frames > 1:
        # could go ahead and combine but warn of inconsistent results
        exit_error(
            "there can only be one  molecular system frame defined in a NEF file"
        )


def offset_chains_in_frames(frames, chain_offsets):
    for frame in frames:
        for chain, offset in chain_offsets.items():
            _offset_residue_numbers(frame, chain, offset)


def _exit_multiple_offsets_for_chain(chain, first_offset, second_offset):
    msg = """
        the chain {chain} has more than one offset/start to use for renumbering

        the first offset/start is {first_offset}
        the second offset/start is {second_offset}
        """
    msg = msg.format(
        input=input, chain=chain, first_offset=first_offset, second_offset=second_offset
    )

    exit_error(msg)


def _get_chain_offset_pairs_or_exit_error(chain_offsets):

    # tee here as we use the chain_offsets generator twice
    chain_offsets_check, chain_offsets = tee(chunks(chain_offsets, 2))

    for chain_offset in chain_offsets_check:
        if len(chain_offset) % 2:
            chain_offset_pairs = []

            for chain_offset in chain_offsets:
                chain_offset_pairs.append(":".join(chain_offset))

            msg = f"""\
                there must be an offset/start for each chain
                i got the following pairs of chain:offset/start: {', '.join(chain_offset_pairs)}
                the last one is missing an offset.
            """
            exit_error(msg)

    result = {}
    seen_chains = set()
    for i, (chain, offset) in enumerate(chain_offsets, start=1):

        _exit_if_offset_isnt_int(chain, offset, i)

        offset = int(offset)

        if chain in seen_chains:
            _exit_multiple_offsets_for_chain(
                chain=chain, first_offset=result[chain], second_offset=offset
            )

        seen_chains.add(chain)
        result[chain] = offset

    return result


def _exit_if_offset_isnt_int(chain, offset, i):
    if not is_int(offset):
        msg = f"""\
                the offset/start {offset} in the {end_with_ordinal(i)} chain offset/start pair: {chain} {offset}
                can't be converted to an int
            """
        exit_error(msg)


def _exit_if_chains_and_offsets_dont_match(chains, offsets):
    num_chains = len(chains)
    num_offsets = len(offsets)
    if num_chains != num_offsets:
        msg = f"""\
            chains and offsets must have the same length i got {num_chains} chains and {num_offsets}
            chains: {','.join(chains)}
            offsets: {','.join([str(offset) for offset in offsets])}
        """
        exit_error(msg)


def _offset_residue_numbers(frame, chain, offset):

    for loop_data in frame.loop_dict.values():
        for tag in loop_data.tags:
            if _tag_is_based_on(tag, SEQUENCE_CODE):
                sequence_index = loop_data.tag_index(tag)

                chain_code_tag = _tag_based_on(tag, SEQUENCE_CODE, CHAIN_CODE)
                chain_code_index = loop_data.tag_index(chain_code_tag)
                for line in loop_data:
                    if chain_code_index is not None and line[chain_code_index] == chain:
                        sequence = line[sequence_index]
                        if is_int(sequence):
                            sequence = int(sequence)
                            sequence = sequence + offset
                            line[sequence_index] = str(sequence)


def _tag_based_on(tag, base, substitute):
    column_code = tag[len(base) :]
    new_tag = f"{substitute}{column_code}"
    return new_tag


def _tag_is_based_on(tag, base):
    return tag.startswith(base)



import json
import traceback
from pathlib import Path
from typing import List

import xmltodict

from nef_pipelines.lib.nef_lib import UNUSED
from nef_pipelines.lib.translation.chem_comp import ChemComp
from nef_pipelines.lib.translation.object_iter import ObjectIter
from nef_pipelines.lib.util import exit_error, nef_pipelines_root


def _items_in(target):
    if isinstance(target, list):
        for index, item in enumerate(target):
            yield index, item
    elif isinstance(target, dict):
        for key, value in target.items():
            yield key, value
    else:
        yield None, target


def _clone_container(targets):
    if isinstance(targets, dict):
        result = {}
    elif isinstance(targets, list):
        result = []
    else:
        result = None
    return result


def _null_mutator(key=None, target=None, path=None):
    return target


def _reporting_mutator(path, target):
    path = [str(item) for item in path]
    target_type = str(type(target)).strip("<>").split()[1].strip("'")
    print(f'{":".join(path)} [{target_type}]: {target}')

    return target


class _Stack(list):
    def push(self, value):
        self.append(value)
        return self


class _Mutator:
    def __init__(self, mutator=_null_mutator):
        self._stack = _Stack()
        self._mutator = mutator

    @property
    def mutator(self):
        return self._mutator

    @mutator.setter
    def mutator(self, mutator):
        self._mutator = mutator

    def process_items(self, target):
        return self._process_items(
            [
                target,
            ]
        )[0]

    def _process_items(self, target):

        result = _clone_container(target)

        for key, target in _items_in(target):
            self._stack.push(key)
            if isinstance(result, list):
                target = self._mutator(self._stack, target)
                processed = self._process_items(target)
                if processed:
                    result.append(processed)
            elif isinstance(result, dict):
                target = self._mutator(self._stack, target)
                processed = self._process_items(target)
                if processed:
                    result[key] = processed
            else:
                processed = self._mutator(self._stack, target)
                if processed.__class__ != target.__class__:
                    processed = self._process_items(processed)

                result = processed
            self._stack.pop()

        return result


class _ReplaceNames:
    def __init__(self, names):
        self._names = names

    def __call__(self, stack, target):
        if isinstance(target, dict):
            for name, replacement in self._names.items():
                if name in target:
                    old = target[name]
                    del target[name]
                    target[replacement] = old

        return target


class _ReplaceAts:
    def __call__(self, stack, target):
        if isinstance(target, dict):
            replacement_names = {}
            for name in target:
                if name.startswith("@"):
                    replacement_names[name] = name[1:]

            for old_name, new_name in replacement_names.items():
                old = target[old_name]
                del target[old_name]
                target[new_name] = old

        return target


class _ReturnBranch:
    def __init__(self, path):
        self._path = path

    def _find_target(self, target):

        for elem in self._path:
            if target is not None and elem in target:
                target = target[elem]

        return target

    def __call__(self, path, target):

        return self._find_target(target)


class _IDSAsStrings:
    @staticmethod
    def _id_as_int(int_id):
        return int(int_id[1:])

    def __call__(self, path, target):

        if isinstance(target, list):
            for index, item in enumerate(target):
                if (
                    isinstance(item, str)
                    and item.startswith("_")
                    and item[1:].isdigit()
                ):
                    target[index] = self._id_as_int(item)
        elif isinstance(target, dict):
            for key, value in target.items():
                if (
                    isinstance(value, str)
                    and value.startswith("_")
                    and value[1:].isdigit()
                ):
                    target[key] = self._id_as_int(value)

        return target


class _ParseIdList:
    @staticmethod
    def _id_as_int(int_id):
        return int(int_id[1:])

    def _as_id_list(self, id_list):

        result = []
        if isinstance(id_list, str):
            for item in id_list.split():
                if item[0] == "_" and item[1:].isdigit():
                    result.append(self._id_as_int(item))
                else:
                    result = None
                    break
        return result

    def __call__(self, path, target):

        if isinstance(target, dict):
            for key, value in target.items():
                as_id_list = self._as_id_list(value)
                if as_id_list:
                    target[key] = as_id_list

        return target


class _ParseBoolean:
    def __call__(self, path, target):

        if isinstance(target, dict):
            for key, value in target.items():
                if value == "true":
                    target[key] = True
                elif value == "false":
                    target[key] = False

        return target


class _FixPseudoLists:
    def __call__(self, path, target):

        if isinstance(target, dict) and len(target) == 1:
            for name in ["IMPL.Line", "IMPL.String"]:
                if isinstance(target, dict) and name in target.keys():
                    value = target[name]
                    if (
                        isinstance(value, str)
                        and value.startswith("[")
                        and value.endswith("]")
                    ):
                        # noinspection PyBroadException
                        try:
                            as_list = eval(value)
                        except Exception:
                            as_list = None

                        if isinstance(as_list, list):
                            target = as_list

        return target


class _FixSingleLineOrString:
    def __call__(self, path, target):

        if isinstance(target, dict) and len(target) == 1:
            keys = list(target.keys())
            for name in ["IMPL.Line", "IMPL.String", "IMPL.Text"]:
                if name in keys:
                    target = target[name]

        return target


class _ConvertNumbers:
    def _as_int(self, target):

        result = None
        try:
            result = int(target)
        except Exception:
            pass

        return result

    def _as_float(self, target):

        result = None
        try:
            result = float(target)
        except Exception:
            pass

        return result

    def __call__(self, path, target):

        if isinstance(target, dict):
            for key, value in target.items():
                if isinstance(value, str):
                    new_value = self._as_int(value)
                    if new_value is not None:
                        target[key] = new_value
                        continue

                    new_value = self._as_float(value)
                    if new_value is not None:
                        target[key] = new_value
                        continue

        return target


class _RemoveNestedTypes:
    def __init__(self, container_type_pairs):
        self._container_type_pairs = container_type_pairs

    def __call__(self, path, target):

        for container, type_or_types in self._container_type_pairs:
            if isinstance(target, dict) and container in target:
                new_contents = []

                types = (
                    [
                        type_or_types,
                    ]
                    if not isinstance(type_or_types, (list, tuple))
                    else type_or_types
                )

                for _type in types:

                    if (
                        isinstance(target[container], dict)
                        and _type in target[container]
                    ):

                        if type(target[container][_type]) == dict:
                            new_contents.append(target[container][_type])
                        else:
                            new_contents.extend(target[container][_type])

                if new_contents:
                    target[container] = new_contents

        return target


class _AddDisciminatorsToApplicationData:
    def __init__(self, types, discriminator="type"):
        self._types = types
        self._discriminator = discriminator

    def __call__(self, path, target):
        for _type in self._types:

            if len(path) >= len(_type) and tuple(path[-2:]) == _type:
                target_type = type(target)
                if target_type == dict:
                    target[self._discriminator] = _type[-1]
                else:
                    for elem in target:
                        elem[self._discriminator] = _type[-1]

        return target


def _atom_info(atoms_by_id, atom):
    atom = atoms_by_id[atom]
    name = atom["name"]

    subtype = ""
    if "subType" in atom:
        subtype = f"[{atom['subType']}]"

    result = f"{name}{subtype}"

    return result


def load_xml_chem_comp_as_json(file_name, verbose=False):

    xml = open(file_name, "rb")
    chemcomp = xmltodict.parse(xml)

    mutator = _Mutator(_ReturnBranch("_StorageUnit:CHEM.StdChemComp".split(":")))
    chemcomp = mutator.process_items(chemcomp)

    name_table = {
        "CHEM.ChemComp.chemAtomSets": "chemAtomSets",
        "CHEM.ChemAtomSet": "ChemAtomSet",
        "CHEM.ChemComp.chemAtoms": "chemAtoms",
        "CHEM.ChemAtom": "ChemAtom",
        "CHEM.ChemComp.chemBonds": "chemBonds",
        "CHEM.ChemBond": "ChemBond",
        "CHEM.ChemComp.chemCompVars": "chemCompVars",
        "CHEM.ChemCompVar": "ChemCompVar",
        "CHEM.ChemComp.chemTorsions": "chemTorsions",
        "CHEM.ChemTorsion": "ChemTorsion",
        "CHEM.ChemComp.linkEnds": "linkEnds",
        "CHEM.LinkEnd": "LinkEnd",
        "CHEM.ChemComp.namingSystems": "namingSystems",
        "CHEM.NamingSystem": "NamingSystem",
        "CHEM.LinkAtom": "LinkAtom",
        "CHEM.ChemCompVar.descriptor": "descriptor",
        "CHEM.ChemComp.commonNames": "commonNames",
        "CHEM.ChemComp.name": "name",
        "CHEM.ChemAtomSet.chemAtoms": "chemAtoms",
        "CHEM.ChemAtomSet.chemAtomSets": "chemAtomSets",
        "CHEM.AbstractChemAtom.chemBonds": "chemBonds",
        "CHEM.AbstractChemAtom.chemCompVars": "chemCompVars",
        "CHEM.AbstractChemAtom.chemTorsions": "chemTorsions",
        "CHEM.ChemAtom.boundLinkEnds": "boundLinkEnds",
        "CHEM.ChemCompVar.stereoSmiles": "stereoSmiles",
        "CHEM.ChemCompVar.chemAtoms": "chemAtoms",
        "CHEM.LinkAtom.boundLinkEnd": "boundLinkEnd",
        "CHEM.ChemAtom.remoteLinkEnds": "remoteLinkEnds",
        "CHEM.LinkAtom.remoteLinkEnd": "remoteLinkEnd",
        "CHEM.ChemBond.chemAtoms": "chemAtoms",
        "CHEM.ChemTorsion.chemAtoms": "chemAtoms",
        "CHEM.ChemTorsion.sysNames": "sysNames",
        "CHEM.LinkEnd.boundLinkAtom": "boundLinkAtom",
        "CHEM.LinkEnd.remoteLinkAtom": "remoteLinkAtom",
        "CHEM.NamingSystem.atomSetVariantSystems": "atomSetVariantSystems",
        "CHEM.NamingSystem.atomSysNames": "atomSysNames",
        "CHEM.AtomSysName": "AtomSysName",
        "CHEM.AtomSysName.sysName": "sysName",
        "CHEM.AtomSysName.altSysName": "altSysName",
        "CHEM.NamingSystem.atomVariantSystems": "atomVariantSystems",
        "CHEM.NamingSystem.chemCompSysName": "chemCompSysNames",
        "CHEM.AtomSysName.altSysNames": "altSysNames",
        "CHEM.NamingSystem.chemCompSysNames": "chemCompSysNames",
        "CHEM.ChemCompSysName": "ChemCompSysName",
        "CHEM.NamingSystem.chemTorsionSysNames": "chemTorsionSysNames",
        "CHEM.ChemTorsionSysName": "ChemTorsionSysName",
        "IMPL.DataObject.applicationData": "applicationData",
        "IMPL.AppDataString": "AppDataString",
        "IMPL.AppDataInt": "AppDataInt",
        "IMPL.AppDataBoolean": "AppDataBoolean",
        "IMPL.AppDataString.value": "value",
        "IMPL.ApplicationData.application": "application",
        "IMPL.ApplicationData.keyword": "keyword",
    }

    replacer = _Mutator(_ReplaceNames(name_table))
    chemcomp = replacer.process_items(chemcomp)

    deatter = _Mutator(_ReplaceAts())
    chemcomp = deatter.process_items(chemcomp)

    replacer = _Mutator(_ReplaceNames({"_ID": "ID"}))
    chemcomp = replacer.process_items(chemcomp)

    id_fixer = _Mutator(_IDSAsStrings())
    chemcomp = id_fixer.process_items(chemcomp)

    id_list_fixer = _Mutator(_ParseIdList())
    chemcomp = id_list_fixer.process_items(chemcomp)

    boolean_fixer = _Mutator(_ParseBoolean())
    chemcomp = boolean_fixer.process_items(chemcomp)

    pseudo_list_fixer = _Mutator(_FixPseudoLists())
    chemcomp = pseudo_list_fixer.process_items(chemcomp)

    pseudo_lines_or_strings = _Mutator(_FixSingleLineOrString())
    chemcomp = pseudo_lines_or_strings.process_items(chemcomp)

    number_converter = _Mutator(_ConvertNumbers())
    chemcomp = number_converter.process_items(chemcomp)

    ATOM_TYPES = {("chemAtoms", "ChemAtom"), ("chemAtoms", "LinkAtom")}

    add_discriminators = _Mutator(_AddDisciminatorsToApplicationData(ATOM_TYPES))
    chemcomp = add_discriminators.process_items(chemcomp)

    APPLICATION_DATA_TYPES = {
        ("applicationData", "AppDataInt"),
        ("applicationData", "AppDataString"),
        ("applicationData", "AppDataBoolean"),
    }

    add_discriminators = _Mutator(
        _AddDisciminatorsToApplicationData(APPLICATION_DATA_TYPES)
    )
    chemcomp = add_discriminators.process_items(chemcomp)

    NESTED_TYPES = (
        ("chemAtoms", ("ChemAtom", "LinkAtom")),
        ("chemBonds", "ChemBond"),
        ("chemTorsions", "ChemTorsion"),
        ("atomSysNames", "AtomSysName"),
        ("chemCompSysNames", "ChemCompSysName"),
        ("chemTorsionSysNames", "ChemTorsionSysName"),
        ("chemCompSysNames", "ChemCompSysName"),
        ("namingSystems", "NamingSystem"),
        ("chemAtomSets", "ChemAtomSet"),
        ("linkEnds", "LinkEnd"),
        ("chemCompVars", "ChemCompVar"),
        ("applicationData", ("AppDataInt", "AppDataString", "AppDataBoolean")),
    )

    remove_nested_types = _Mutator(_RemoveNestedTypes(NESTED_TYPES))
    chemcomp_data = remove_nested_types.process_items(chemcomp)

    if "CHEM.NonStdChemComp" in chemcomp_data:
        raise ValueError(
            f"Error: Non standard chem comps are not currently supported {file_name}"
        )

    return ChemComp(**chemcomp_data)


def convert_xml_chem_comps(
    chem_comp_paths: List[Path], out_dir: Path = None, force=False
):
    for path in chem_comp_paths:
        # if path.parts[-1] == 'DNA+I+msd_ccpnRef_2007-12-11-10-13-17_00002.xml':
        _convert_xml_chem_comp(path, out_dir, force=force)


def type_children(target, parent):

    result = ()
    if hasattr(target, "__origin__"):
        result = result = tuple(
            [ObjectIter.Entry(target, i, arg) for i, arg in enumerate(target.__args__)]
        )

    elif isinstance(target, (list, tuple)):
        result = tuple(
            [ObjectIter.Entry(target, i, elem) for i, elem in enumerate(target)]
        )

    elif isinstance(target, dict):
        result = tuple(
            [ObjectIter.Entry(target, key, value) for key, value in target.items()]
        )

    return result


def _convert_xml_chem_comp(chem_comp_path: Path, out_dir: Path = None, force=False):
    # print(chem_comp_path)
    file_iter = chem_comp_path.iterdir()
    if chem_comp_path.is_file():
        file_iter = [chem_comp_path]

    if out_dir:
        out_dir = Path(_chem_comp_root_dir(), *out_dir)
        if not out_dir.exists():
            out_dir.mkdir()
        if not out_dir.is_dir():
            exit_error(f"output directory must be a directory {out_dir} isn't")

    for file_path in file_iter:
        if file_path.suffix == ".xml":

            ok = False
            try:

                chem_comp = load_xml_chem_comp_as_json(file_path)
                json_chem_comp = file_path.with_suffix(".json")
                if out_dir:
                    json_chem_comp = out_dir / json_chem_comp.parts[-1]

                if json_chem_comp.exists() and not force:
                    exit_error(
                        f"file {json_chem_comp} exists to overwite use then force option"
                    )

                raw_json = chem_comp.json()
                with open(json_chem_comp, "w") as fh:
                    json_dict = json.loads(raw_json)
                    json_formatted = json.dumps(json_dict, indent=4)
                    fh.write(json_formatted)

                ok = True

            except Exception as e:
                caught = e
                if (
                    "Error: Non standard chem comps are not currently supported"
                    not in caught.args[0]
                ):
                    print(traceback.format_exc())
                    print(f"Error: loading {file_path} failed because {e}")

            if ok:
                print(f"OK: converted {file_path} to json")
                print(f"    new file  {json_chem_comp}")
            else:
                if (
                    "Error: Non standard chem comps are not currently supported"
                    not in caught.args[0]
                ):
                    print(f"Error: loading {file_path} failed")


CHEM_COMP_REL_PATH = ["nef_pipelines", "data", "chem_comp"]
JSON_CHEM_COMP_SUB_PATH = (".",)
XML_CHEM_COMP_SUB_PATH = ("xml",)


def _chem_comp_root_dir():
    return Path(nef_pipelines_root(), *CHEM_COMP_REL_PATH)


def find_chem_comps(rel_path=JSON_CHEM_COMP_SUB_PATH, extension="json") -> List[Path]:

    result = []
    for file_path in Path(_chem_comp_root_dir(), *rel_path).iterdir():
        if file_path.suffix == f".{extension}":
            result.append(file_path)

    return result


MOL_TYPES = set()
CHEM_COMPS = {}


# rough costs
# read + parse std chem comps 0.218489s with v2 expected to be 0.054-0.004s typically 0.012s
# just read chem comps 0.010838s
# without linking 0.160787s (linking oh 0.058s ie 26%)
def load_chem_comps():
    chem_comp_paths = find_chem_comps()

    for chem_comp_path in chem_comp_paths:
        with open(chem_comp_path, "r") as f:
            chemcomp_data = json.load(f)
            chem_comp = ChemComp(**chemcomp_data)
            mol_type = chem_comp.molType
            MOL_TYPES.add(mol_type.upper())
            key = (
                chem_comp.code3Letter
                if chem_comp.code3Letter != UNUSED
                else chem_comp.ccpCode
            )
            CHEM_COMPS[key] = chem_comp


if __name__ == "__main__":

    # xml_chem_comp_paths = find_chem_comps(rel_path=XML_CHEM_COMP_SUB_PATH,extension='xml')
    #
    # convert_xml_chem_comps(xml_chem_comp_paths, JSON_CHEM_COMP_SUB_PATH, force=True)
    #
    # print(CHEM_COMPS)

    load_chem_comps()
    # print(CHEM_COMPS)
    #
    # for type,residue in CHEM_COMPS:
    #     print(type, residue)
    #     name = f'{type.lower()}_{residue.lower()}_test.json'
    #     path = Path(_chem_comp_root_dir()) / name
    #     chem_comp = CHEM_COMPS[type,residue]
    #     chem_comp_json = chem_comp.json()
    #     print(path)
    #     with open(path,'w') as fh:
    #         fh.write(json.dumps(json.loads(chem_comp_json), indent=4))
    # print()

    print(CHEM_COMPS["PROTEIN", "ILE"].namingSystems[0].atomReference.name)



import sys

from nef_pipelines import nef_app

if nef_app:
    # noinspection PyUnusedLocal
    @nef_app.app.command()
    def sink():
        """- read the current stream and don't write anything"""
        if not sys.stdin.isatty():
            for line in sys.stdin:
                pass



import string
from pathlib import Path

import pytest

from nef_pipelines.lib.util import (
    STDOUT,
    exit_if_file_has_bytes_and_no_force,
    fnmatch_one_of,
    strip_characters_left,
    strip_characters_right,
)


@pytest.mark.parametrize(
    "input, expected",
    [
        ("", ("", "")),
        ("AAAA", ("", "AAAA")),
        ("123AAAA", ("123", "AAAA")),
    ],
)
def test_strip_characters_left(input, expected):

    result = strip_characters_left(input, string.digits)

    assert result == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        ("", ("", "")),
        ("AAAA", ("AAAA", "")),
        ("AAAA123", ("AAAA", "123")),
    ],
)
def test_strip_characters_right(input, expected):

    result = strip_characters_right(input, string.digits)
    assert result == expected


@pytest.mark.parametrize(
    "target, patterns,  result",
    [
        ("a", "a b a".split(), True),
        ("a", "d e f".split(), False),
        ("abc", "a*c e f".split(), True),
        ("abc", "abc e f".split(), True),
    ],
)
def test_fnmatch_one_of(target, patterns, result):
    assert fnmatch_one_of(target, patterns) == result


@pytest.mark.parametrize(
    "file_name, force, only_touch, is_exception_raised",
    [
        (STDOUT, False, None, False),
        (STDOUT, True, None, False),
        ("test.txt", True, True, False),
        ("test.txt", False, True, False),
        ("test.txt", True, False, False),
        ("test.txt", False, False, True),
        (None, True, None, False),
        (None, False, None, False),
    ],
)
def test_exit_if_file_has_bytes_and_no_force(
    file_name, force, only_touch, is_exception_raised, tmpdir
):

    if file_name == STDOUT:
        path = STDOUT
    elif file_name:
        path = Path(tmpdir) / file_name
    else:
        path = Path(tmpdir) / "non_existent.txt"

    if file_name and path != STDOUT:
        if only_touch:
            path.touch()
        else:
            with path.open("w") as f:
                f.write("test")

    exception_raised = False
    try:
        exit_if_file_has_bytes_and_no_force(path, force)
    except SystemExit:
        exception_raised = True

    assert exception_raised == is_exception_raised



import sys
from pathlib import Path
from typing import List, Tuple

import typer
from pynmrstar import Entry
from tabulate import tabulate

from nef_pipelines.lib.nef_lib import (
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames_by_name,
)
from nef_pipelines.lib.shift_lib import (
    frames_to_assigned_and_unassigned_shift_lists,
    shifts_to_chains,
)
from nef_pipelines.lib.util import STDIN, STDOUT, exit_if_file_has_bytes_and_no_force
from nef_pipelines.transcoders.mars import export_app
from nef_pipelines.transcoders.mars.util import (
    _exit_if_no_frames_selected,
    _filter_shifts_by_chain,
    _select_target_chain_from_sequence_if_not_defined,
)

app = typer.Typer()


# noinspection PyUnusedLocal
@export_app.command()
def fixed(
    input_path: Path = typer.Option(
        STDIN,
        "--input",
        "-i",
        metavar="|INPUT|",
        help="input to read NEF data from [stdin = -]",
    ),
    shift_frame_selectors: List[str] = typer.Argument(
        None,
        help="selector for the shift restraint frames to use, can be called multiple times and include wild cards",
    ),
    target_chain: str = typer.Option(
        None,
        "-c",
        "--chain",
        help="chain to export [# and @ and the 1st chain code in the project if the is only one]",
        metavar="<CHAIN-CODE>",
    ),
    output_file: str = typer.Option(
        None,
        "-o",
        "--out",
        help="file name to output to [default <entry_id>_fix_ass.tab] for stdout use -",
        metavar="<MARS-FIXED-ASSIGNMENT>",
    ),
    force: bool = typer.Option(
        False,
        "-f",
        "--force",
        help="force overwrite of output file if it exists and isn't empty",
    ),
):
    """- convert chemical shifts and restraints to fixed assignments"""

    if not shift_frame_selectors:
        shift_frame_selectors = ["*"]

    entry = read_entry_from_file_or_stdin_or_exit_error(input_path)

    output_file = (
        f"{entry.entry_id}_fix_ass.tab" if output_file is None else output_file
    )
    output_file = Path(output_file)

    entry = pipe(entry, shift_frame_selectors, target_chain, output_file, force)

    if entry:
        print(entry)


def pipe(
    entry: Entry,
    shift_frame_selectors: List[str],
    target_chain: str,
    output_file: Path,
    force: bool,
) -> List[List[Tuple[str, str]]]:

    frames = entry.get_saveframes_by_category("nef_chemical_shift_list")
    frames = select_frames_by_name(frames, shift_frame_selectors)

    _exit_if_no_frames_selected(frames)

    assigned_shifts, unassigned_shifts = frames_to_assigned_and_unassigned_shift_lists(
        frames
    )

    shift_chains = shifts_to_chains([*assigned_shifts, *unassigned_shifts])

    target_chain = _select_target_chain_from_sequence_if_not_defined(
        target_chain, shift_chains
    )

    table = _build_assigned_shifts(assigned_shifts, target_chain)

    exit_if_file_has_bytes_and_no_force(output_file, force)

    file_h = sys.stdout if output_file == STDOUT else open(output_file, "w")

    print(tabulate(table, tablefmt="plain"), file=file_h)

    if output_file != STDOUT:
        file_h.close()

    result = None
    if output_file != STDOUT:
        result = entry

    return result


def _build_assigned_shifts(assigned_shifts, target_chain):
    result = []
    assigned_shifts = _filter_shifts_by_chain(assigned_shifts, target_chain)
    assigned_residues = {shift.atom.residue for shift in assigned_shifts}
    for assigned_residue in sorted(assigned_residues):
        residue_name = assigned_residue.residue_name
        sequence_code = assigned_residue.sequence_code
        assigned_residue_name = f"AR_{sequence_code}_{residue_name}"
        result.append([assigned_residue_name, str(sequence_code)])

    return result



from collections import Counter
from dataclasses import replace
from pathlib import Path
from textwrap import dedent, indent
from typing import List

import typer
from fyeah import f
from pynmrstar import Entry

from nef_pipelines.lib.nef_frames_lib import (
    EXPERIMENT_CLASSIFICATION,
    EXPERIMENT_TYPE,
    SHIFT_LIST_FRAME_CATEGORY,
)
from nef_pipelines.lib.nef_lib import (
    UNUSED,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames_by_name,
)
from nef_pipelines.lib.peak_lib import peaks_to_frame
from nef_pipelines.lib.shift_lib import nef_frames_to_shifts
from nef_pipelines.lib.spectra_lib import (
    EXPERIMENT_CLASSIFICATION_TO_SYNONYM,
    EXPERIMENT_GROUPS,
    EXPERIMENT_INFO,
    FAKE_SPECTROMETER_FREQUENCY_600,
    SPECTRUM_TYPE_TO_CLASSIFICATION,
    ExperimentType,
    PeakInfo,
)
from nef_pipelines.lib.structures import NewPeak, ShiftData
from nef_pipelines.lib.util import (
    FOUR_SPACES,
    STDIN,
    exit_error,
    flatten,
    is_int,
    parse_comma_separated_options,
    strings_to_tabulated_terminal_sensitive,
)
from nef_pipelines.tools.simulate import simulate_app

SPECTRUM_NAME_TEMPLATE = "synthetic_{spectrum}"

SHIFT_FRAMES_HELP = """\
    the names of the shift frames to use, this can be a comma separated list of name or the option can be called
    called multiple times. Wild cards are allowed. If no match is found wild cards are checked as well unless
    --exact is set. If no shift frames are selected the default chemical shift list frame is used
"""

SPECTRA_HELP = f"""
    the names of the spectra to create, this can be a comma separated list of options or
    called multiple times. Possible values are: {', '.join(ExperimentType)}
"""
NAME_TEMPLATE_HELP = """
    the name template for the new peak lists the placeholder {spectrum} will get replaced with the spectrum type
"""


# TODO: add a flag to list spectra and groups...
@simulate_app.command()
def peaks(
    shift_frame_selectors: List[str] = typer.Option(
        None, "--shift-frames", help=SHIFT_FRAMES_HELP
    ),
    exact: bool = typer.Option(
        False, help="if set frames are selected by exact matches"
    ),
    name_template: str = typer.Option(SPECTRUM_NAME_TEMPLATE, help=NAME_TEMPLATE_HELP),
    spectrometer_frequency: float = typer.Option(FAKE_SPECTROMETER_FREQUENCY_600),
    input: Path = typer.Option(
        STDIN,
        "--in",
        "-i",
        metavar="|INPUT|",
        help="input to read NEF data from [stdin = -]",
    ),
    spectra: List[str] = typer.Argument(None, help=SPECTRA_HELP),
):
    """-  make a set of peaks for an hsqc, 13C direct detect or triple resonance spectrum from a list of shifts
    [alpha for non ccpn peak lists and 13C detect]"""

    spectra = [spectrum.replace("-", "_") for spectrum in spectra]
    if not spectra:
        spectra = [
            ExperimentType.N_HSQC,
        ]

    shift_frame_selectors = (
        [
            "nef_chemical_shift_list_default",
        ]
        if not shift_frame_selectors
        else parse_comma_separated_options(shift_frame_selectors)
    )

    shift_frames = parse_comma_separated_options(shift_frame_selectors)

    spectra = set(parse_comma_separated_options(spectra))

    spectra = _strs_to_experiment_types_or_exit_error(spectra)

    spectra = _update_spectra_with_groups(spectra)

    # TODO add a check that we have some shift frames

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    entry = pipe(
        entry, shift_frames, exact, spectra, name_template, spectrometer_frequency
    )

    print(entry)


def pipe(
    entry: Entry,
    shift_frame_selectors: List[str],
    exact_frame_selectors: bool,
    spectra: List[ExperimentType],
    name_template_string: str,
    spectrometer_frequency: float = FAKE_SPECTROMETER_FREQUENCY_600,
) -> Entry:

    all_shift_list_frames = entry.get_saveframes_by_category(SHIFT_LIST_FRAME_CATEGORY)

    selected_shift_list_frames = select_frames_by_name(
        all_shift_list_frames, shift_frame_selectors, exact=exact_frame_selectors
    )

    shifts = nef_frames_to_shifts(selected_shift_list_frames)

    for spectrum in spectra:

        if spectrum in EXPERIMENT_INFO:
            spectrum_info = EXPERIMENT_INFO[spectrum]
        elif spectrum_info.lower() in EXPERIMENT_INFO:
            spectrum_info = EXPERIMENT_INFO[spectrum.lower()]

        peaks = _make_spectrum(shifts, spectrum_info)

        if len(peaks) == 0:
            continue

        frame_name = f(name_template_string)

        dimensions = [
            {"axis_code": dimension} for dimension in spectrum_info.dimensions
        ]

        # TODO add info about frames for errors
        if spectrum not in SPECTRUM_TYPE_TO_CLASSIFICATION:
            _exit_bad_spectrum_type(spectrum, SPECTRUM_TYPE_TO_CLASSIFICATION.keys())

        spectrum_classification = SPECTRUM_TYPE_TO_CLASSIFICATION[spectrum]
        extra_tags = {
            EXPERIMENT_CLASSIFICATION: spectrum_classification,
            EXPERIMENT_TYPE: EXPERIMENT_CLASSIFICATION_TO_SYNONYM[
                spectrum_classification
            ],
        }
        frame = peaks_to_frame(
            peaks, dimensions, spectrometer_frequency, frame_name, extra_tags=extra_tags
        )

        entry.add_saveframe(frame)

    return entry


def _sign_of_value(value):
    result = ""
    result = "+" if value > 0 else result
    result = "-" if value < 0 else result
    return result


def _average_atom_set_shifts(atom_set_shifts):
    shifts_by_atom = {}
    for atom_set_shift in atom_set_shifts:
        residue = atom_set_shift.atom.residue
        atom_key = (
            residue.chain_code,
            residue.sequence_code,
            residue.offset,
            atom_set_shift.atom.atom_name,
        )
        shifts_by_atom.setdefault(atom_key, []).append(atom_set_shift)

    result = []
    for key, values in shifts_by_atom.items():
        if len(values) == 1:
            result.append(values[0])
            continue
        else:
            value = sum([shift.value for shift in values]) / len(values)
            result.append(ShiftData(values[0].atom, round(value, 3), UNUSED))

    return result


def _make_spectrum(
    shifts: List[ShiftData], info: PeakInfo, height: int = 1_000_000
) -> List[NewPeak]:

    # note this is not very sophisticated and doesn't cover side-chains... we ought to port the ccpn code

    updated_shifts = []

    for shift in shifts:

        sequence_code = str(shift.atom.residue.sequence_code)
        prefix = sequence_code[0] if sequence_code.startswith("@") else ""
        sequence_code = sequence_code.lstrip("@")

        # TODO not sure this is the best way to do it, separators are only relevant if the sequence code starts with @
        found_separator = False
        for separator in ["-", "+"]:
            if sequence_code.startswith(separator):
                sequence_sign = sequence_code[0]
                sequence_code = sequence_code[1:]
            else:
                sequence_sign = ""

            if separator in sequence_code and not sequence_code[0] == separator:
                sequence_code_fields = sequence_code.split(separator)
                if len(sequence_code_fields) != 2:
                    updated_shifts.append(shift)
                    continue

                sequence_code, offset = sequence_code_fields

                if sequence_code[0] == "@":
                    sequence_code = sequence_code[1:]
                    prefix = "@"

                sequence_code_int = (
                    int(f"{sequence_sign}{sequence_code}")
                    if is_int(sequence_code)
                    else None
                )
                offset_int = int(offset) if is_int(offset) else None

                if not sequence_code_int and not offset_int:
                    updated_shifts.append(shift)
                    continue

                if separator == "+":
                    sequence_code_int = sequence_code_int + offset_int
                elif separator == "-":
                    sequence_code_int = sequence_code_int - offset_int

                residue = shift.atom.residue
                residue = replace(residue, sequence_code=sequence_code_int)
                residue = replace(residue, offset=offset_int)
                if prefix:
                    residue = replace(residue, sequence_code_prefix=prefix)

                atom = replace(shift.atom, residue=residue)
                shift = replace(shift, atom=atom)
                updated_shifts.append(shift)
                found_separator = True

        if not found_separator:
            sequence_code = str(shift.atom.residue.sequence_code)
            if sequence_code.startswith("@"):
                prefix = "@"
                sequence_code = sequence_code[1:]
                if is_int(sequence_code):
                    sequence_code = int(sequence_code)
                    residue = shift.atom.residue
                    residue = replace(residue, sequence_code=sequence_code)
                    residue = replace(residue, sequence_code_prefix=prefix)
                    atom = replace(shift.atom, residue=residue)
                    shift = replace(shift, atom=atom)
            updated_shifts.append(shift)

    shifts = updated_shifts

    relevant_shifts = []

    for atom_name in flatten(info.atoms):
        for shift in shifts:
            if shift.atom.atom_name.upper() in atom_name:
                relevant_shifts.append(shift)

    shifts_by_chain_prefix_and_residue = {}

    for shift in relevant_shifts:
        residue = shift.atom.residue.sequence_code
        chain = shift.atom.residue.chain_code
        prefix = shift.atom.residue.sequence_code_prefix
        shifts_by_chain_prefix_and_residue.setdefault(
            (chain, prefix, residue), []
        ).append(shift)

    chains_prefix_residues = list(shifts_by_chain_prefix_and_residue.keys())

    peaks = []

    for i, (chain, prefix, residue_number) in enumerate(sorted(chains_prefix_residues)):

        if not is_int(residue_number):
            continue

        for atom_set, sign in zip(info.atom_sets, info.atom_set_signs):

            atom_set_shifts = []
            for j, (offset, atom_name_required) in enumerate(atom_set):
                targeted_residue = residue_number + offset

                per_residue_shifts = shifts_by_chain_prefix_and_residue.get(
                    (chain, prefix, targeted_residue), []
                )
                for k, shift in enumerate(per_residue_shifts):
                    if shift.atom.atom_name == atom_name_required:
                        atom_set_shifts.append(shift)

            atom_set_shifts = _average_atom_set_shifts(atom_set_shifts)

            if len(atom_set_shifts) > len(atom_set):
                continue

            for i, shift in enumerate(atom_set_shifts):
                sequence_code = shift.atom.residue.sequence_code
                sequence_code_prefix = shift.atom.residue.sequence_code_prefix
                sequence_code_offset = shift.atom.residue.offset

                if sequence_code_offset and sequence_code_prefix == "@":
                    sequence_code = int(sequence_code) + sequence_code_offset
                    sign_of_offset = _sign_of_value(-sequence_code_offset)
                    sequence_code = (
                        f"{sequence_code}{sign_of_offset}{sequence_code_offset}"
                    )

                if sequence_code_prefix:
                    sequence_code = f"{sequence_code_prefix}{sequence_code}"
                    residue = shift.atom.residue
                    residue = replace(
                        residue, sequence_code=sequence_code, sequence_code_prefix=""
                    )
                    atom = replace(shift.atom, residue=residue)
                    shift = replace(shift, atom=atom)
                    atom_set_shifts[i] = shift

            if len(atom_set_shifts) == len(info.dimensions):
                new_peak = NewPeak(
                    atom_set_shifts, height=height * sign, volume=height * sign
                )
                peaks.append(new_peak)

    return peaks


def _strs_to_experiment_types_or_exit_error(spectra):
    result = []
    spectrum_names = set([experiment.value for experiment in ExperimentType])
    for spectum in spectra:

        found = False
        if spectum in spectrum_names:
            result.append(ExperimentType[spectum])
            found = True

        experiment_types_upper = [
            experiment_type.name.upper() for experiment_type in ExperimentType
        ]
        upper_experiments_to_experiments = {
            experiment_type.name.upper(): experiment_type
            for experiment_type in ExperimentType
        }
        counts = Counter(experiment_types_upper)

        if not found:
            spectrum_upper = spectum.upper()
            if spectrum_upper in experiment_types_upper and counts[spectrum_upper] == 1:
                result.append(upper_experiments_to_experiments[spectrum_upper])
                found = True
            if spectrum_upper in experiment_types_upper and counts[spectrum_upper] > 1:
                degenerate_spectra = [
                    experiment_type.name
                    for experiment_type in ExperimentType
                    if experiment_type.upper() == spectrum_upper
                ]
                _exit_degnerate_spectrum_type(
                    spectum, list(ExperimentType), degenerate_spectra
                )

        if not found:
            _exit_bad_spectrum_type(spectum, list(ExperimentType))

    return result


def _update_spectra_with_groups(spectra):

    spectra = set(spectra)
    to_update = []
    for experiment_type in spectra:
        if experiment_type in EXPERIMENT_GROUPS:
            to_update.append(experiment_type)

    for experiment_type in to_update:
        spectra.update(EXPERIMENT_GROUPS[experiment_type])
        spectra.remove(experiment_type)

    return spectra


def _exit_bad_spectrum_type(spectrum, known_spectra):
    msg = """
        the spectrum type {spectrum} is not know, available spectrum types are

        {spectrum_types}
    """
    msg = dedent(msg)

    spectrum_types = strings_to_tabulated_terminal_sensitive(known_spectra)
    spectrum_types = indent(spectrum_types, FOUR_SPACES)

    exit_error(f(msg))


def _exit_degnerate_spectrum_type(spectrum, known_spectra, degenerate_spectra):
    degenerate_spectra = ", ".join(degenerate_spectra)

    spectrum_types = strings_to_tabulated_terminal_sensitive(known_spectra)
    spectrum_types = indent(spectrum_types, FOUR_SPACES)

    msg = f"""
            the spectrum type {spectrum} doesn't exactly match a spectrum type in the list below
            and has more than possible match when case is ignored [{degenerate_spectra}]

            {spectrum_types}
        """
    msg = dedent(msg)

    exit_error(f(msg))
